#ifndef _HAPLOTYPETEST_TYPEDEFS_
#define _HAPLOTYPETEST_TYPEDEFS_
#include <stdio.h>

#ifndef _HAPLOTYPETEST_PARAMETERS_
#include "define.h"
#endif

/* Type definitions for atom */
typedef unsigned int  haplotype;
typedef unsigned char allele;

/* Allele name */
typedef struct allele_name {
  char          Allele_Name[CHAR_IN_ALLELE_NAME+1];   /* allele name */
} Allele_Name;

/* Allele name at one locus */
typedef struct locus_data {
  char          Allele_Name[2][CHAR_IN_ALLELE_NAME+1];   /* allele name */
  int           number_of_allele;   /* number of allele */
  int           count[3];           /* allele count. 3rd for missing */
  double        Allele_Frequency;   /* major allele frequency for each locus */
  int           genotype_count[3];  /* genotype count */
  int           filtered;           /* flag of filtered locus */
  double        HWE_check;          /* statistics for HWE test */
  double        p_value;            /* p-value of HWE test */
  double        PIC;                /* PIC value */
  double        heterozygosity;     /* heterozyosity of the locus */
  double        hetero_variance;    /* variance of heterozyosity */
} Locus_Data;

/* Allele at one locus */
typedef struct locus_code {
  allele        allele_code;        /* allele code */
} Locus_Code;

/* Allele at all locus */
typedef struct subject_genotype {
  Locus_Code   *Locus;              /* list of allele code for all loci */
  int           packed_offset;      /* offset in haplotype inference table*/
  int           phenotype;          /* phenotype flag */
  double        QTL;                /* QTL value */
  double        QTL_value[2];       /* QTL value input and converted to Normal distribution */
  char          Name[CHAR_IN_SUBJECT_NAME+1];
} Subject_Genotype;

/* Input genotype */
typedef struct genotype_data {
  int           Number_of_Loci;     /* number of loci in genotype file */
  int           Number_of_Subjects; /* number of subjects in genotype file */
  int           Number_of_Effective_Subjects; /* number of subjects used for estimation */
  int           Number_of_Unknown;  /* number of subjects which don't have phenotype data */
  Locus_Data   *Locus_List;         /* list of allele name for all loci */
  Subject_Genotype *Subjects;       /* list of genotype for each subject */
} Genotype_Data;

/* Diplotype configuration */
typedef struct diplotype {
  haplotype      diplotype[2];    /* haplotype combination */
} Diplotype;

typedef struct diploclass {       /* flags for genetic model */
  unsigned char  h_flag[2];
  unsigned char  d_flag;
  unsigned char  p_flag;
} DiploClass;

/* Packed genotype */
typedef struct genotype_pack {
  haplotype      Genotype_Code[2]; /* packed form of genotype */
  haplotype      Missing[2];       /* missing data flags */
  int            phenotype;
  double         weight;           /* weight for the genotype */
  int            number_of_combi;  /* number of haplotype combination */
  Diplotype     *combi;            /* haplotype combination array */
  DiploClass    *class;            /* flags for genetic model */
  double        *penetrance;       /* pointer to penetrance */
} Genotype_Pack;

/* Haplotype inference table */
typedef struct estimate_set {
  int            number_of_set;   /* number of different genotypes */
  int            df_for_test;     /* degree of freedom for test */
  int            df_for_test_Q1;     /* degree of freedom for test */
  int            df_for_test_s;     /* degree of freedom for test */
  double         normal;          /* normalization factor */
  double         normal_given;    /* normalization factor for phenotype-known subjects */
  double         p_plus[3];       /* penetrance       : no target haplotype, one target haplotype, two target haplotypes */
  double         p_minus[3];      /* 1.0 - penetrance : no target haplotype, one target haplotype, two target haplotypes */
  double         mu[3];           /* mu with no target haplotype, one target haplotype, two target haplotypes */
  double         SD;              /* overall standard deviation */
  double         SD_c[3];         /* standard deviation with no target haplotype, one target haplotype, two target haplotypes */
  double         contingency[3][2]; /* corresponding contingency table */
  int            model_type;      /* trait inheritance model */
  haplotype      target;          /* target haplotype */
  haplotype      target_mask;     /* target haplotype mask */
  int            select_loci;     /* number of loci to use */
  int            df;              /* degree of freedom */
  int            total_combi;     /* total number of haplotype combination */
  int            max_haplo_combi; /* maximum number of haplotype combination */
  int            number_of_haplo; /* number of possible haplotype */
  int            nonzero;         /* number of non zero frequency haplotype */
  int           *loci;            /* loci to use */
  double        *f_allele;        /* major allele frequency for selected loci */
  Genotype_Pack *geno_info;       /* genotype for each set */
  Diplotype     *combi;           /* haplotype combination array */
  DiploClass    *class;           /* flags for genetic model */
  haplotype     *haplo;           /* possible haplotype list */
  haplotype     *order;           /* order of haplotype frequency */
  double        *f;               /* haplotype frequency */
} Estimate_Set;

/* Haplotype subset information */
typedef struct haplotype_subset {
  haplotype     target;           /* target haplotype */
  haplotype     mask;             /* target haplotype mask */
  haplotype     member;           /* member of haplotype set */
  double        statistic;        /* statistic of test */
  double        p_value;          /* P value of chi-square test */
  double        p_value_p;        /* P value of permutation test */
  double        result[4];        /* result of test */
  double        interval[4][2];   /* confidence interval of result */
  double        odds_ratio;       /* odds ratio for case-control */
  double        odds_interval[2]; /* confidence interval of odds ratio */
} Haplotype_Subset;

/* Sliding window test result */
typedef struct window_result {
  double        statistic;        /* statistic of test */
  double        p_value;          /* P value of chi-square test */
  double        result[4];        /* result of test */
  double        interval[4][2];   /* confidence interval of result */
  double        odds_ratio;       /* odds ratio for case-control */
  double        odds_interval[2]; /* confidence interval of odds ratio */
  char         *haplotype_name;
} Window_Result;

/* Haplotype set table */
typedef struct haplotype_set {
  int           number_of_htSNP;  /* number of htSNPs detected */
  int           number_of_subset; /* number of haplotype set */
  haplotype     loci_mask;        /* htSNP position in packed format */
  int          *htSNP_loci;       /* htSNP position */
  Haplotype_Subset *subset;       /* pointer to haplotype set */
} Haplotype_Set;

/* Option for calculation */
typedef struct option_data {
  int           genotype_form;    /* genotype format */
  int           haplotype;        /* haplotype inference option */
  int           init_type;        /* selection of initial distribution for EM */
  int           freq_table;       /* popultion haplotype frequency table flag */
  int           max_loci;         /* maximum region length */
  int           overlap;          /* number of overlapped loci */
  int           maximize;         /* number of restart to search maximum lod */
  int           maximize_f;       /* extra message control of restart */
  double        SNP_rejection;    /* rejection frequency of SNP allele */
  int           missing_loci;     /* rejection number of loci */
  int           pairLD;           /* LD measure based on pair-wise calculation */
  int           combiLD;          /* haplotype for all combination of loci */
  int           width;            /* window size for haplotype heterozygosity */
  int           bootstrap;        /* number for resampling for bootstrap */
  int           permutation;      /* number for generating samples for permutation test */
  int           diplo_out;        /* extra output control for diplotype */
  double        criteria;         /* criteria for EM iteration */
  double        s_level;          /* significance level */
  int           message;          /* extra message control */
  int           select_loci;      /* number of loci to use */
  int          *loci;             /* loci to use */
  char         *target;           /* target haplotype pattern */
  int           number_of_model;  /* number of trait inheritance model to be tested */
  int           model_type[4];    /* trait inheritance model */
  int           switch_p;         /* complement of phenotype */
  int           penetrance;       /* inference with penetrance */
  double        p_plus[3];        /* penetrance for no target haplotype, one target haplotype, two target haplotypes */
  int           QTL;              /* QTL analysis */
  int           phenotype_format; /* phenotype file format */
  int           htSNP;            /* find htSNPs */
  int           select_htSNP;     /* number of htSNPs specified by user */
  int          *loci_htSNP;       /* loci of htSNPs specified by user */
  int           haplotype_set;    /* flag for test of SNP anchored haplotype set */
  double        cut_frequency;    /* frequency criteria for test of SNP anchored haplotype set */
  int           convertQTL;       /* convert QTL phenotype distribution to Normal distribution */
  int           window_flag;      /* flag for window calculation */
  int           window_size;      /* window size for window calculation */
  int           msg_flag;         /* flag for output */
  int           outlier;          /* flag for outlier calculation */
  int           conf_flag;        /* confidence interval calculation flag */
  double        confidence;       /* confidence interval criteria */
  int           check_ID;
} Option_Data;

/* File information */
typedef struct file_status_data {
  char          option_name[FILENAME_MAX];
  char          genotype_name[FILENAME_MAX];
  char          phenotype_name[FILENAME_MAX];
  char          haplotype_freqL_name[FILENAME_MAX];
  int           interactive;    /* interactive mode */
  FILE         *option_fp;
  int           genotype_stat;
  FILE         *genotype_fp;
  int           phenotype_stat;
  FILE         *phenotype_fp;
  int           output_stat;
  FILE         *output_list;
} File_Status_data;

#endif
