#include "haplotypetest.h"

extern void *mem_alloc(size_t, const char *);
extern int   error_message(char *);
extern int   warning_message(char *);
extern int   error_check();

int  init_option(Option_Data *);
int  get_option(char *, Option_Data *);
#ifdef DEBUG
int  dump_option(Option_Data *);
#endif

int  init_option(Option_Data *option)
{
  /* Initialize option data */
  option->genotype_form    = 0; /* original format is default */
  option->haplotype        = 0; /* Haplotype is not infered */
  option->max_loci         = MAX_LOCI_REGION;
  option->overlap          = 0;
  option->maximize         = 0; /* No restart */
  option->maximize_f       = 0;
  option->SNP_rejection    = SNP_REJECTION;
  option->missing_loci     = 0;
  option->pairLD           = 0;
  option->combiLD          = 0;
  option->width            = 0;
  option->init_type        = 0; /* Uniform is default */
  option->bootstrap        = 0;
  option->permutation      = 0;
  option->diplo_out        = 0;
  option->criteria         = CRITERIA;
  option->s_level          = SIGNIFICANCE;
  option->message          = 0;
  option->select_loci      = 0;
  option->loci             = NULL;
  option->target           = NULL;
  option->number_of_model  = 1;
  option->model_type[0]    = DOMINANT_MODEL;
  option->model_type[1]    = RECESSIVE_MODEL;
  option->model_type[2]    = GENOTYPE_MODEL;
  option->model_type[3]    = OVER_DOMINANT_MODEL;
  option->switch_p         = 0;
  option->penetrance       = 0;
  option->p_plus[0]        = 0.9;
  option->p_plus[1]        = 0.9;
  option->p_plus[2]        = 0.9;
  option->QTL              = 0;
  option->phenotype_format = 1;
  option->htSNP            = 0;
  option->select_htSNP     = 0;
  option->loci_htSNP       = NULL;
  option->haplotype_set    = 0;
  option->cut_frequency    = CUT_FREQUENCY;
  option->convertQTL       = 0;
  option->window_flag      = 0;
  option->window_size      = WINDOW_SIZE;
  option->msg_flag         = 1;
  option->outlier          = 0;
  option->conf_flag        = 0;
  option->confidence       = CONFIDENCE;
  option->check_ID         = 0;

  return 0;
}

int  get_option(char *option_file, Option_Data *option)
{
  int          tag_max     = 30;
  char        *tag_type[]  = {"haplotype_heterozygosity",
			      "target_haplotype",
			      "haplotype",
			      "pairwiseLD",
			      "combinationLD",
			      "initialize",
			      "bootstrap",
			      "permutation",
			      "confidence_interval",
			      "diplotype_output",
			      "EM_iteration_criteria",
			      "significance_level",
			      "select_htSNP",
			      "select_loci",
			      "maximum_LRC_search",
			      "SNP_rejection",
			      "genotype_format",
			      "reject_missing",
			      "initial_penetrance",
			      "switch_penetrance",
			      "extra_message",
                              "QTL",
			      "htSNP",
                              "test_of_haplotype_set",
                              "mode_of_inheritance",
                              "convert_Normal_distribution",
                              "phenotype_format",
			      "outlier",
                              "window",
                              "check_ID"};
  int          effective_len[] = { 14,  6,  8, 10, 13, 10,  9, 11, 19,  9,
				   12, 12, 12,  6, 11, 10, 15, 14, 11, 10,
				   13,  3,  5, 21, 15, 14,  9,  7,  6,  8};
  char         delm        = '=';
  /*  char         init_max    = 4;*/
  char         init_max    = 3;
  char        *init_type[] = {"uniform","equilibrium","random","mixed"};
  char         format_max    = 2;
  char        *format_type[] = {"original","prefile"};
  char         phenotype_format_max    = 2;
  char        *phenotype_format_type[] = {"with_ID","no_ID"};
  char         inheritance_max     = 4;
  char        *inheritance_type[] = {"dominant","recessive","additive","over_dominant"};
  int          i, type, not_comment;
  char        *real_buffer, *read_buffer, *tag, *end, *par;
  int          item, m, n, *ptr;
  double       dval;

  /* Open option file */
  if (strlen(option_file) == 0) {
    File_Status->interactive   = 1;
  }
  else {
    File_Status->option_fp = fopen(option_file,"r");
  }

  if (File_Status->option_fp == NULL) {
    error_message("Option file open error.");
  }
  if (error_check()) return;

  /* Copy option file name */
  strcpy(File_Status->option_name, option_file);
  end = strstr(File_Status->option_name,".dat");
  if (end != NULL) {
    *end = '\0';
  }

  /* Read option data */
  printf("Option data echo\n==========\n");

  real_buffer = (char *)mem_alloc(MAX_BUFFER_LENGTH*sizeof(char),"Read buffer");
  while(fgets(real_buffer, MAX_BUFFER_LENGTH-1, File_Status->option_fp) != NULL) {
    read_buffer = real_buffer;

    /* Delete \r and \n */
    end = strchr(read_buffer, CARRIAGE_RETURN);
    if (end == NULL) {
      end = strchr(read_buffer, LINE_FEED);
    }
    if (end != NULL) {
      *end = STRING_TERMINATOR;
    }
    printf("%s\n",read_buffer);

    /* First:  check comment "#" */
    not_comment = 1;
    end = strchr(read_buffer, COMMENT_CHAR);
    if (end != NULL) {
      *end = '\0';
      not_comment = 0;
    }
    if (strlen(read_buffer) == 0) {
      not_comment = 0;
    }

    /* Second: check proceeding blank */
    while (*read_buffer != STRING_TERMINATOR) {
      if ((*read_buffer != DELIMITER_SPACE) && (*read_buffer != DELIMITER_TAB)) {
	break;
      }
      read_buffer++;
    }

    type = -1;
    for (i=0; i<tag_max; i++) {
      if (strncmp(read_buffer, tag_type[i], effective_len[i]) == 0) {
	type = i;
	break;
      }
    }

    end = strchr(read_buffer, delm);

    switch (type) {
    case 0: /* haplotype_heterozygosity */
      if (end != NULL) {
	end++;
	item = sscanf(end,"%d",&m);
	if (item == 1) {
	  if (m > 0) {
	    option->width = m;
	  }
	  else {
	    error_message("Window size to calculate heterozygosity is an erroneous value.");
	  }
	}
	else {
	  error_message("Haplotype heterozygosity calculation is specified, but no window size is entered.");
	}
      }
      else {
	error_message("Haplotype heterozygosity calculation is specified, but no window size is entered.");
      }
      break;
    case 1: /* target_haplotype */
      if (end != NULL) {
	end++;

	par = end;
	while ((*par != '\0') &&
	       ((*par != LINE_FEED) && (*par != CARRIAGE_RETURN))) {
	  par++;
	}
	*par = '\0';

	item = par-end+1;
	if (item > 0) {
	  if (option->target != NULL) {
	    free(option->target);
	  }
	  if (option->QTL == 0) {
	    option->penetrance = 1;
	  }
	  option->target = (char *)mem_alloc(item*sizeof(char),"Target haplotype string");

	  strcpy(option->target, end);
	}
	else {
	  error_message("Target haplotype is specified, but no haplotype is entered.");
	}
      }
      else {
	error_message("Target haplotype is specified, but no haplotype is entered.");
      }
      break;
    case 2: /* haplotype */
      option->haplotype = 1;
      if (end != NULL) {
	end++;
	item = sscanf(end,"%d,%d",&m,&n);

	switch (item) {
	case 2:
	  if ((n > -1) && (n <= MAX_LOCI_OVERLAP) && (m > n)) {
	    option->overlap   = n;
	  }
	  else {
	    error_message("Overlap size on dividing loci is an erroneous value or too large.");
	  }
	case 1:
	  if ((m >  1) && (m <= MAX_LOCI_REGION)   ) {
	    option->max_loci  = m;
	  }
	  else {
	    error_message("Loci divided size is an erroneous value.");
	  }
	  break;
	default:
	  break;
	}
      }
      break;
    case 3: /* pairwiseLD */
      option->pairLD  = 1;
      break;
    case 4: /* combinationLD */
      option->combiLD = 1;
      break;
    case 5: /* initialize */
      if (end != NULL) {
	option->init_type = -1;
	end++;
	for (i=0; i<init_max; i++) {
	  tag = strstr(end, init_type[i]);

	  if (tag != NULL) {
	    option->init_type = i;
	    break;
	  }
	}
	if (option->init_type == -1) {
	  error_message("Initialize option is invalid.");
	}
      }
      else {
	option->init_type = 0;
      }
      break;
    case 6: /* bootstrap */
      if (end != NULL) {
	end++;
	item = sscanf(end,"%d",&m);
	if (item == 1) {
	  if (m > 0) {
	    option->bootstrap = m;
	  }
	  else {
	    error_message("Number of samples for bootstrap is an erroneous value.");
	  }
	}
	else {
	  error_message("Bootstrap calculation is specified, but number of samples is not entered.");
	}
      }
      else {
	error_message("Bootstrap calculation is specified, but number of samples is not entered.");
      }
      break;
    case 7: /* permutation */
      if (end != NULL) {
	end++;
	item = sscanf(end,"%d",&m);
	if (item == 1) {
	  if (m > 0) {
	    option->permutation = m;
	  }
	  else {
	    error_message("Number of samples for permutation test is an erroneous value.");
	  }
	}
	else {
	  error_message("Permutation test is specified, but number of samples is not entered.");
	}
      }
      else {
	option->permutation = -1;
      }
      break;
    case 8: /* confidence_interval */
      option->conf_flag = 1;
      if (end != NULL) {
	end++;
	item = sscanf(end,"%lf",&dval);
	if (item == 1) {
	  if ((dval <= 1.0) && (dval >= 0.0)) {
	    option->confidence = dval;
	  }
	  else {
	    error_message("The value of confidence interval criteria must be between 0.0 and 1.0.");
	  }
	}
      }
      break;
    case 9: /* diplotype_output */
      option->diplo_out = 1;
      break;
    case 10: /* EM_iteration_criteria */
      if (end != NULL) {
	end++;
	item = sscanf(end,"%lf",&dval);
	if (item == 1) {
	  option->criteria = dval + 1.0;
	}
	else {
	  error_message("EM iteration criteria is specified, but value of criteria is not entered.");
	}
      }
      else {
	error_message("EM iteration criteria is specified, but value of criteria is not entered.");
      }
      break;
    case 11: /* significance_level */
      if (end != NULL) {
	end++;
	item = sscanf(end,"%lf",&dval);
	if (item == 1) {
	  if ((dval <= 1.0) && (dval >= 0.0)) {
	    option->s_level = dval;
	  }
	  else {
	    error_message("The value of significance level must be between 0.0 and 1.0.");
	  }
	}
	else {
	  error_message("Significance level is specified, but value of limit is not entered.");
	}
      }
      else {
	error_message("Significance level is specified, but value of limit is not entered.");
      }
      break;
    case 12: /* select_htSNP */
      if (end != NULL) {
	par = end;
	item = 0;
	while ((*par != '\0') &&
	       ((*par != LINE_FEED) && (*par != CARRIAGE_RETURN))) {
	  par++;
	  if (*par == ',') {
	    item++;
	  }
	}

	item++; option->select_htSNP = item;
	if (option->loci_htSNP != NULL) {
	  free(option->loci_htSNP);
	}
	option->loci_htSNP = (int *)mem_alloc(item*sizeof(int),"Selected htSNP loci number array");

	ptr = option->loci_htSNP;
	while (end != NULL) {
	  end++;
	  item = sscanf(end,"%d",ptr);
	  ptr++;
	  end = strchr(end, ',');
	}

	option->htSNP         = 1;
	option->haplotype_set = 1;
	if (option->QTL == 0 && option->penetrance == 0) {
	  option->penetrance = 1;
	}
      }
      else {
	error_message("htSNP loci selection is specified, but no loci number is entered.");
      }
      break;
    case 13: /* select_loci */
      if (end != NULL) {
	par = end;
	item = 0;
	while ((*par != '\0') &&
	       ((*par != LINE_FEED) && (*par != CARRIAGE_RETURN))) {
	  par++;
	  if (*par == ',') {
	    item++;
	  }
	}

	item++; option->select_loci = item;
	if (option->loci != NULL) {
	  free(option->loci);
	}
	option->loci = (int *)mem_alloc(item*sizeof(int),"Selected loci number array");

	ptr = option->loci;
	while (end != NULL) {
	  end++;
	  item = sscanf(end,"%d",ptr);
	  ptr++;
	  end = strchr(end, ',');
	}
      }
      else {
	error_message("Loci selection is specified, but no loci number is entered.");
      }
      break;
    case 14: /* maximum_LRC_search */
      if (end != NULL) {
	end++;
	item = sscanf(end,"%d,%d",&m,&n);

	switch (item) {
	case 2:
	  option->maximize_f = n;
	case 1:
	  if (m > 0) {
	    option->maximize = m;
	  }
	  else {
	    error_message("Number of loop for maximum LRC search is an erroneous value.");
	  }
	  break;
	default:
	  error_message("Number of loop for maximum LRC search is an erroneous value.");
	}
      }
      else {
	error_message("Maximum LRC search option is specified, but loop number is not entered.");
      }
      break;
    case 15: /* SNP_rejection */
      if (end != NULL) {
	end++;
	item = sscanf(end,"%lf",&dval);
	if (item == 1) {
	  if ((dval < 0.5) && (dval > 0.0)) {
	    option->SNP_rejection = dval;
	  }
	  else {
	    error_message("The value of SNP rejection criteria must be between 0.0 and 0.5.");
	  }
	}
	else {
	  error_message("SNP rejection criteria is specified, but no value of criteria is entered.");
	}
      }
      else {
	error_message("SNP rejection criteria is specified, but no value of criteria is entered.");
      }
      break;
    case 16: /* genotype format */
      if (end != NULL) {
	option->genotype_form = -1;
	end++;
	for (i=0; i<init_max; i++) {
	  tag = strstr(end, format_type[i]);

	  if (tag != NULL) {
	    option->genotype_form = i;
	    break;
	  }
	}
	if (option->genotype_form == -1) {
	  error_message("Genotype file format option is invalid.");
	}
      }
      else {
	option->genotype_form = 0;
      }
      break;
    case 17: /* reject_missing */
      if (end != NULL) {
	end++;
	item = sscanf(end,"%d",&m);
	if (item == 1) {
	  option->missing_loci = m;
	}
	else {
	  error_message("Missing loci rejection criteria is specified, but no value of criteria is entered.");
	}
      }
      else {
	error_message("Missing loci rejection criteria is specified, but no value of criteria is entered.");
      }
      break;
    case 18: /* initial_penetrance */
      if (end != NULL) {
	end++;
	item = sscanf(end,"%lf,%lf,%lf",&option->p_plus[0],&option->p_plus[1],&option->p_plus[2]);
	if (item < 2) {
	  error_message("Initial penetrances for target haplotype and the others must be specified.");
	}
      }
      else {
	error_message("Initial penetrances are specified, but no value of penetrance is entered.");
      }
      break;
    case 19: /* switch_penetrance */
      option->switch_p = 1;
      break;
    case 20: /* extra_message */
      option->message = 1;
      break;
    case 21: /* QTL */
      option->QTL        = 1;
      option->penetrance = 0;
      if (end != NULL) {
	option->phenotype_format = -1;
	end++;
	for (i=0; i<phenotype_format_max; i++) {
	  tag = strstr(end, phenotype_format_type[i]);

	  if (tag != NULL) {
	    option->phenotype_format = i;
	    break;
	  }
	}
	if (option->phenotype_format == -1) {
	  error_message("QTL file format option is invalid.");
	}
      }
      break;
    case 22: /* htSNP */
      option->htSNP      = 1;
      if (end != NULL) {
	end++;
	item = sscanf(end,"%lf",&dval);
	if (item == 1) {
	  if ((dval < 0.5) && (dval >= 0.0)) {
	    option->cut_frequency = dval;
	  }
	  else {
	    error_message("The value of frequency criteria for haplotype must be between 0.0 and 0.5.");
	  }
	}
      }
      break;
    case 23: /* test_of_haplotype_set */
      option->htSNP         = 1;
      option->haplotype_set = 1;
      if (option->QTL == 0 && option->penetrance == 0) {
	option->penetrance = 1;
      }
      if (end != NULL) {
	end++;
	item = sscanf(end,"%lf",&dval);
	if (item == 1) {
	  if ((dval < 0.5) && (dval >= 0.0)) {
	    option->cut_frequency = dval;
	  }
	  else {
	    error_message("The value of frequency criteria for haplotype must be between 0.0 and 0.5.");
	  }
	}
      }
      break;
    case 24: /* mode_of_inheritance */
      option->number_of_model = 0;
      if (end != NULL) {
	end++;
	while ((tag = strsep(&end, ",")) != NULL) {
	  for (i=0; i<inheritance_max; i++) {
	    if (strcmp(tag, inheritance_type[i]) == 0) {
	      if (option->number_of_model < 4) {
		option->model_type[option->number_of_model] = i+1;
		option->number_of_model++;
	      }
	      else {
		error_message("Total number of mode of inheritance exceeds 4.");
	      }
	      break;
	    }
	  }
	  if (i == inheritance_max) {
	    sprintf(error_message_buffer,"Mode of inheritance, %s, is invalid.",tag);
	    error_message(error_message_buffer);
	  }
	}
      }
      if (option->number_of_model == 0) {
	error_message("Mode of inheritance is invalid.");
      }
      break;
    case 25: /* convert_Normal_distribution */
      option->convertQTL = 1;
      break;
    case 26: /* phenotype_format */
      if (end != NULL) {
	option->phenotype_format = -1;
	end++;
	for (i=0; i<phenotype_format_max; i++) {
	  tag = strstr(end, phenotype_format_type[i]);

	  if (tag != NULL) {
	    option->phenotype_format = i;
	    break;
	  }
	}
	if (option->phenotype_format == -1) {
	  error_message("phenotype file format option is invalid.");
	}
      }
      else {
	option->phenotype_format = 1;
      }
      break;
    case 27: /* outlier */
      option->outlier = 1;
      break;
    case 28: /* window */
      option->window_flag = 1;
      if (end != NULL) {
	end++;
	item = sscanf(end,"%d",&m);
	if (item == 1) {
	  if (m > 0) {
	    option->window_size = m;
	  }
	  else {
	    error_message("Window size is an erroneous value.");
	  }
	}
      }
      break;
    case 29: /* check_ID */
      option->check_ID = 1;
      break;
    default:
      if (not_comment) {
	error_message("Unknown option data.");
      }
    }
  }

  printf("==========\n");

  fclose(File_Status->option_fp);
  File_Status->option_fp = NULL;
  free(real_buffer);

  if (option->window_flag) {
    option->select_htSNP  = 0;
    option->htSNP         = 1;
    option->haplotype_set = 1;
  }
  if (option->haplotype_set || option->window_flag) {
    option->bootstrap = 0;
    option->diplo_out = 0;
    printf("\nTest of haplotype set or window calculation is specified.\nBootstrap, diplotype output, and target haplotype options are ignored.\n");
  }

#ifdef DEBUG
  dump_option(option);
#endif

  return 0;
}

#ifdef DEBUG
int  dump_option(Option_Data *option)
{
  int l;
  /* Debug output of option data */
  printf("# option->genotype_form   = %d\n",option->genotype_form);
  printf("# option->haplotype       = %d\n",option->haplotype);
  printf("# option->init_type       = %d\n",option->init_type);
  printf("# option->max_loci        = %d\n",option->max_loci);
  printf("# option->overlap         = %d\n",option->overlap);
  printf("# option->maximize        = %d\n",option->maximize);
  printf("# option->maximize_f      = %d\n",option->maximize_f);
  printf("# option->SNP_rejection   = %f\n",option->SNP_rejection);
  printf("# option->missing_loci    = %d\n",option->missing_loci);
  printf("# option->pairLD          = %d\n",option->pairLD);
  printf("# option->combiLD         = %d\n",option->combiLD);
  printf("# option->width           = %d\n",option->width);
  printf("# option->bootstrap       = %d\n",option->bootstrap);
  printf("# option->permutation     = %d\n",option->permutation);
  printf("# option->diplo_out       = %d\n",option->diplo_out);
  printf("# option->criteria        = %f\n",option->criteria);
  printf("# option->s_level         = %f\n",option->s_level);
  printf("# option->message         = %d\n",option->message);
  printf("# option->select_loci     = %d\n",option->select_loci);
  printf("# option->loci            =");
  for (l=0; l<option->select_loci; l++) {
    printf(" %d",*(option->loci + l));
  }
  printf("\n");
  printf("# option->target          = %s\n",option->target);
  printf("# option->number_of_model = %d\n",option->number_of_model);
  printf("# option->model_type      =");
  for (l=0; l<option->number_of_model; l++) {
    printf(" %d",option->model_type[l]);
  }
  printf("\n");
  printf("# option->switch_p        = %d\n",option->switch_p);
  printf("# option->penetrance      = %d\n",option->penetrance);
  printf("# option->p_plus[0]       = %f\n",option->p_plus[0]);
  printf("# option->p_plus[1]       = %f\n",option->p_plus[1]);
  printf("# option->p_plus[2]       = %f\n",option->p_plus[2]);
  printf("# option->QTL             = %d\n",option->QTL);
  printf("# option->phenotype_format= %d\n",option->phenotype_format);
  printf("# option->htSNP           = %d\n",option->htSNP);
  printf("# option->select_htSNP    = %d\n",option->select_htSNP);
  printf("# option->loci_htSNP      =");
  for (l=0; l<option->select_htSNP; l++) {
    printf(" %d",*(option->loci_htSNP + l));
  }
  printf("\n");
  printf("# option->haplotype_set   = %d\n",option->haplotype_set);
  printf("# option->cut_frequency   = %f\n",option->cut_frequency);
  printf("# option->convertQTL      = %d\n",option->convertQTL);
  printf("# option->window_flag     = %d\n",option->window_flag);
  printf("# option->window_size     = %d\n",option->window_size);
  printf("# option->msg_flag        = %d\n",option->msg_flag);
  printf("# option->conf_flag       = %d\n",option->conf_flag);
  printf("# option->confidence      = %f\n",option->confidence);
  printf("# option->check_ID        = %d\n",option->check_ID);

  return 0;
}
#endif
