#include "haplotypetest.h"

/* Utility function */
extern void          *mem_alloc(size_t, const char *);
extern void           hpsort_reverse_index_of_double_array(int, double *, unsigned int *);
extern void           hpsort_index_of_double_array(int, double *, unsigned int *);

/* Error handler */
extern int            error_check();
extern int            error_message(char *);
extern int            warning_message(char *);

/* Define functions */
int            Outlier_by_genotype_set(Estimate_Set *, Option_Data *, Genotype_Data *, double *, double ());
int            Outlier_by_subject(Estimate_Set *, Option_Data *, Genotype_Data *, double *, double ());

int            Outlier_by_genotype_set(Estimate_Set  *table,
				       Option_Data   *option,
				       Genotype_Data *genotype,
				       double        *f,
				       double         (*likelihood_function)())
{
  /* Local variables */
  int           RC, s, i, p, cum, inner_flag, outer_flag;
  int          *num_subj;
  unsigned int *order;
  int           pos[4];
  double        Q[3], IQR, inner, outer;

  /* Likelihood */
  double        likelihood, *l_subject;

  RC = 0;

  /* Allocate subject likelihood array */
  l_subject = (double       *)mem_alloc(table->number_of_set*sizeof(double),"Likelihood array for each subject");
  num_subj  = (int          *)mem_alloc(table->number_of_set*sizeof(int),"Array for number of subject");
  order     = (unsigned int *)mem_alloc(table->number_of_set*sizeof(int),"Sort Array for subject likelihood");

  /* Set number of subjects for each genotype set */
  for (s=0; s<table->number_of_set; s++) {
    *(num_subj + s) = (int)(table->geno_info + s)->weight;
  }

  /* Get subject likelihood */
  likelihood = (*likelihood_function)(table, f, l_subject);

  /* Sort subject likelihood */
  hpsort_index_of_double_array(table->number_of_set, l_subject, order);

  /* Get cumulative number of subjects and quartile */
  pos[0] = (int)((double)genotype->Number_of_Subjects * 0.25 + 0.5);
  pos[1] = (int)((double)genotype->Number_of_Subjects * 0.50 + 0.5);
  pos[2] = (int)((double)genotype->Number_of_Subjects * 0.75 + 0.5);
  pos[3] = genotype->Number_of_Subjects + 1;
  cum = 0;
  p   = 0;

  printf("\nDistribution of log10 likelihood of each subject\n");
  printf("  order  log10 likelihood     N  sum N\n");
  for (s=0; s<table->number_of_set; s++) {
    cum                      = cum + *(num_subj + *(order+s));
    printf("%7d %17.5e %5d %6d",s+1,*(l_subject+*(order+s)),*(num_subj + *(order+s)),cum);
    while (pos[p] <= cum) {
      pos[p] = s;
      Q[p]   = *(l_subject+*(order+s));
      if      (p == 0) {
	printf("  1st quartile");
      }
      else if (p == 1) {
	printf("  Median");
      }
      else {
	printf("  3rd quartile");
      }
      p++;
    }
    printf("\n");
  }

  IQR   = Q[2] - Q[0];
  printf("\n  IQR (3rd quartile - 1st quartile) = %12.5e\n",IQR);
  /* Outlier exceeds lower bound */
  inner = Q[0]  - 1.5 * IQR;
  outer = inner - 1.5 * IQR;
  printf("  Inner fence of 1st quartile       = %12.5e\n",inner);
  printf("  Outer fence of 1st quartile       = %12.5e\n",outer);
  inner_flag = 1;
  outer_flag = 1;
  for (s=0; s<table->number_of_set; s++) {
    if (*(l_subject+*(order+s)) < inner) {
      RC = 1;
      if (*(l_subject+*(order+s)) < outer) {
	if (outer_flag) {
	  outer_flag = 0;
	  printf("  Far out of 1st quartile\n");
	}
      }
      else {
	if (inner_flag) {
	  inner_flag = 0;
	  printf("  Outside data of 1st quartile\n");
	}
      }
      for (i=0; i<genotype->Number_of_Subjects; i++) {
	if ((genotype->Subjects + i)->packed_offset == *(order+s)) {
	  printf("    Subject '%s', log10 likelihood = %12.5e\n",(genotype->Subjects + i)->Name,*(l_subject+*(order+s)));
	}
      }
    }
    else {
      break;
    }
  }

  /* Outlier exceeds upper bound */
  inner = Q[2]  + 1.5 * IQR;
  outer = inner + 1.5 * IQR;
  printf("  Inner fence of 3rd quartile       = %12.5e\n",inner);
  printf("  Outer fence of 3rd quartile       = %12.5e\n",outer);
  inner_flag = 1;
  outer_flag = 1;
  if (inner < 0.0) {
    for (s=0; s<table->number_of_set; s++) {
      if (*(l_subject+*(order+s)) > inner) {
	RC = 1;
	if (*(l_subject+*(order+s)) > outer) {
	  if (outer_flag) {
	    outer_flag = 0;
	    printf("  Far out of 3st quartile\n");
	  }
	}
	else {
	  if (inner_flag) {
	    inner_flag = 0;
	    printf("  Outside data of 3st quartile\n");
	  }
	}
	for (i=0; i<genotype->Number_of_Subjects; i++) {
	  if ((genotype->Subjects + i)->packed_offset == *(order+s)) {
	    printf("    Subject '%s', log10 likelihood = %12.5e\n",(genotype->Subjects + i)->Name,*(l_subject+*(order+s)));
	  }
	}
      }
      else {
	break;
      }
    }
  }

  free(order);
  free(num_subj);
  free(l_subject);

  return RC;
}

int            Outlier_by_subject(Estimate_Set  *table,
				  Option_Data   *option,
				  Genotype_Data *genotype,
				  double        *f,
				  double         (*likelihood_function)())
{
  /* Local variables */
  int           RC, s, i, p, cum, inner_flag, outer_flag;
  unsigned int *order;
  int           pos[4];
  double        Q[3], IQR, inner, outer;

  /* Likelihood */
  double        likelihood, *l_subject;

  RC = 0;

  /* Allocate subject likelihood array */
  l_subject = (double       *)mem_alloc(genotype->Number_of_Subjects*sizeof(double),"Likelihood array for each subject");
  order     = (unsigned int *)mem_alloc(genotype->Number_of_Subjects*sizeof(int),"Sort Array for subject likelihood");

  /* Get subject likelihood */
  likelihood = (*likelihood_function)(table, genotype, f, l_subject);

  /* Sort subject likelihood */
  hpsort_index_of_double_array(genotype->Number_of_Subjects, l_subject, order);

  /* Get cumulative number of subjects and quartile */
  pos[0] = (int)((double)genotype->Number_of_Subjects * 0.25 + 0.5);
  pos[1] = (int)((double)genotype->Number_of_Subjects * 0.50 + 0.5);
  pos[2] = (int)((double)genotype->Number_of_Subjects * 0.75 + 0.5);
  pos[3] = genotype->Number_of_Subjects + 1;
  cum = 0;
  p   = 0;

  printf("\nDistribution of log10 likelihood of each subject\n");
  printf("  order  log10 likelihood     N  sum N\n");
  for (s=0; s<genotype->Number_of_Subjects; s++) {
    printf("%7d %17.5e %5d %6d",s+1,*(l_subject+*(order+s)),1,s+1);
    while (pos[p] == s) {
      Q[p]   = *(l_subject+*(order+s));
      if      (p == 0) {
	printf("  1st quartile");
      }
      else if (p == 1) {
	printf("  Median");
      }
      else {
	printf("  3rd quartile");
      }
      p++;
    }
    printf("\n");
  }

  IQR   = Q[2] - Q[0];
  printf("\n  IQR (3rd quartile - 1st quartile) = %12.5e\n",IQR);
  /* Outlier exceeds lower bound */
  inner = Q[0]  - 1.5 * IQR;
  outer = inner - 1.5 * IQR;
  printf("  Inner fence of 1st quartile       = %12.5e\n",inner);
  printf("  Outer fence of 1st quartile       = %12.5e\n",outer);
  inner_flag = 1;
  outer_flag = 1;
  for (s=0; s<genotype->Number_of_Subjects; s++) {
    if (*(l_subject+*(order+s)) < inner) {
      RC = 1;
      if (*(l_subject+*(order+s)) < outer) {
	if (outer_flag) {
	  outer_flag = 0;
	  printf("  Far out of 1st quartile\n");
	}
      }
      else {
	if (inner_flag) {
	  inner_flag = 0;
	  printf("  Outside data of 1st quartile\n");
	}
      }
      printf("    Subject '%s', log10 likelihood = %12.5e\n",(genotype->Subjects+*(order+s))->Name,*(l_subject+*(order+s)));
    }
    else {
      break;
    }
  }

  /* Outlier exceeds upper bound */
  inner = Q[2]  + 1.5 * IQR;
  outer = inner + 1.5 * IQR;
  printf("  Inner fence of 3rd quartile       = %12.5e\n",inner);
  printf("  Outer fence of 3rd quartile       = %12.5e\n",outer);
  inner_flag = 1;
  outer_flag = 1;
  if (inner < 0.0) {
    for (s=0; s<genotype->Number_of_Subjects; s++) {
      if (*(l_subject+*(order+s)) > inner) {
	RC = 1;
	if (*(l_subject+*(order+s)) > outer) {
	  if (outer_flag) {
	    outer_flag = 0;
	    printf("  Far out of 3st quartile\n");
	  }
	}
	else {
	  if (inner_flag) {
	    inner_flag = 0;
	    printf("  Outside data of 3st quartile\n");
	  }
	}
	printf("    Subject '%s', log10 likelihood = %12.5e\n",(genotype->Subjects+*(order+s))->Name,*(l_subject+*(order+s)));
      }
      else {
	break;
      }
    }
  }

  free(order);
  free(l_subject);

  return RC;
}
