#include "haplotypetest.h"

/* Utility function */
extern void          *mem_alloc(size_t, const char *);
extern void           hpsort_reverse_index_of_double_array(int, double *, unsigned int *);

/* Error handler */
extern int            error_check();
extern int            error_message(char *);
extern int            warning_message(char *);

/* Diplotype probability */
extern double         diplotype_probability(int, Diplotype *, double *, double *);
extern double         diplotype_probability_with_penetrance(int, Diplotype *, DiploClass *, int, Estimate_Set *, double *, double *);
extern double         diplotype_probability_with_QTL(int, Diplotype *, DiploClass *, double, Estimate_Set *, double *, double *);

/* P-value for chi square test */
extern double         calc_p_value_by_chi_square(int, double);

/* Define functions */
int            print_p_value_of_LRC(int, double);
int            print_contingency_table_for_penetrance(Estimate_Set *, double, double [2], double, int);
int            print_count_for_QTL(Estimate_Set *);
int            print_allele_table(Estimate_Set *, Genotype_Data *, double *);
int            print_haplotype_frequency(Estimate_Set *, double *, Genotype_Data *);
int            print_haplotype_frequency_with_parameters(Estimate_Set *, Estimate_Set *, double *, Genotype_Data *);
int            print_penetrances(Estimate_Set *, double [][2], double, int);
int            print_QTL_parameters(Estimate_Set *, double [][2], double, int);
int            print_haplotype_frequency_by_bootstrap(Estimate_Set *, double *, double *, Genotype_Data *);
int            print_diplotype_profile(Estimate_Set *, Genotype_Data *);
int            print_diplotype_profile_with_penetrance(Estimate_Set *, Genotype_Data *);
int            print_diplotype_profile_with_QTL(Estimate_Set *, Genotype_Data *);
int            print_test_of_haplotype_set_with_penetrance(Estimate_Set *, Genotype_Data *, Haplotype_Set *, unsigned int *, double, int);
int            print_test_of_haplotype_set_with_QTL(Estimate_Set *, Genotype_Data *, Haplotype_Set *, unsigned int *, double, int);
int            print_haplotype(haplotype, Estimate_Set *, Genotype_Data *);
int            fprint_haplotype(FILE *, haplotype, Estimate_Set *, Genotype_Data *);
int            print_haplotype_with_mask(haplotype, haplotype, Estimate_Set *, Genotype_Data *);
int            fprint_haplotype_with_mask(FILE *, haplotype, haplotype, Estimate_Set *, Genotype_Data *);
int            sprint_haplotype_with_mask(char *, haplotype, haplotype, Estimate_Set *, Genotype_Data *);
char          *sprint_odds_ratio(double);
char          *sprint_odds_ratio_7(double);

int            print_p_value_of_LRC(int           df,
				    double        LRC)
{
  double    conv = 2.0 * log(10.0);
  double    s, p_value;

  if (LRC < 0) {
    s = 0.0;
  }
  else {
    s = conv * LRC;
  }
  if (df > 0) {
    p_value = calc_p_value_by_chi_square(df, s);
  }
  else {
    p_value = 1.0;
  }
  printf("\nStatistic for chi-square test by LRC = %.4f\n",s);
  printf("                   degree_of_freedom = %d\n",df);
  printf("                             p_value = %.16e\n",p_value);

  return 0;
}

int            print_contingency_table_for_penetrance(Estimate_Set *table,
						      double        odds_ratio,
						      double        odds_interval[2],
						      double        confidence,
						      int           conf_flag)
{
  int    h, p;
  double sum0, sum1, sum, conf;

  sum0 = 0.0;
  sum1 = 0.0;
  if      (table->df_for_test == 1 || table->df_for_test == 2) {
    printf("\nCorresponding contengency table\n");
    printf("    Diplotype Configuration   Phenotype 0   Phenotype 1\n");
  }
  if (table->model_type == GENOTYPE_MODEL) {
    sum = table->contingency[0][0] + table->contingency[0][1];
    sum0 = sum0 + table->contingency[0][0];
    sum1 = sum1 + table->contingency[0][1];
    printf("   with no target haplotype   %11.2f   %11.2f   %11.2f\n",table->contingency[0][0],table->contingency[0][1],sum);
    sum = table->contingency[1][0] + table->contingency[1][1];
    sum0 = sum0 + table->contingency[1][0];
    sum1 = sum1 + table->contingency[1][1];
    printf("   with 1 target haplotype    %11.2f   %11.2f   %11.2f\n",table->contingency[1][0],table->contingency[1][1],sum);
    sum = table->contingency[2][0] + table->contingency[2][1];
    sum0 = sum0 + table->contingency[2][0];
    sum1 = sum1 + table->contingency[2][1];
    printf("   with 2 target haplotypes   %11.2f   %11.2f   %11.2f\n",table->contingency[2][0],table->contingency[2][1],sum);
    sum  = sum0 + sum1;
    printf("                              %11.2f   %11.2f   %11.2f\n",sum0,sum1,sum);
  }
  else {
    sum  = table->contingency[0][0] + table->contingency[0][1];
    sum0 = sum0 + table->contingency[0][0];
    sum1 = sum1 + table->contingency[0][1];
    printf("   with no target haplotype   %11.2f   %11.2f   %11.2f\n",table->contingency[0][0],table->contingency[0][1],sum);
    sum  = table->contingency[1][0] + table->contingency[1][1];
    sum0 = sum0 + table->contingency[1][0];
    sum1 = sum1 + table->contingency[1][1];
    printf("      with target haplotype   %11.2f   %11.2f   %11.2f\n",table->contingency[1][0],table->contingency[1][1],sum);
    sum  = sum0 + sum1;
    printf("                              %11.2f   %11.2f   %11.2f\n",sum0,sum1,sum);

    if (conf_flag) {
      conf = confidence * 100.0;
      printf("  Odds ratio for case-control study = %s  %3.0f%% CI: ",sprint_odds_ratio(odds_ratio),conf);
      printf("%s -",sprint_odds_ratio(odds_interval[0]));
      printf("%s\n",sprint_odds_ratio(odds_interval[1]));
    }
    else {
      printf("  Odds ratio for case-control study = %s\n",sprint_odds_ratio(odds_ratio));
    }
  }

  return 0;
}

int            print_count_for_QTL(Estimate_Set *table)
{
  int     h;
  double  sum;

  sum = 0.0;
  printf("\nCount of diplotype configuration\n");
  printf("    Diplotype Configuration       count\n");
  if (table->model_type == GENOTYPE_MODEL) {
    sum = sum + table->contingency[0][0];
    printf("   with no target haplotype   %9.2f\n",table->contingency[0][0]);
    sum = sum + table->contingency[1][0];
    printf("   with 1 target haplotype    %9.2f\n",table->contingency[1][0]);
    sum = sum + table->contingency[2][0];
    printf("   with 2 target haplotypes   %9.2f\n",table->contingency[2][0]);
    printf("                              %9.2f\n",sum);
  }
  else {
    sum = sum + table->contingency[0][0];
    printf("   with no target haplotype   %9.2f\n",table->contingency[0][0]);
    sum = sum + table->contingency[1][0];
    printf("      with target haplotype   %9.2f\n",table->contingency[1][0]);
    printf("                              %9.2f\n",sum);
  }

  return 0;
}

int            print_allele_table(Estimate_Set  *table,
				  Genotype_Data *genotype,
				  double        *f_allele)
{
  /* Local variables */
  int    l, i;

  printf("\nAllele frequency for each locus after haplotype inference\n");
  printf("  locus    name  input   estimate  name  input   estimate\n");
  for (i=0; i<table->select_loci; i++) {
    l = *(table->loci + i);
    printf("%7d   %5s  %6.4f  %6.4f   %5s  %6.4f  %6.4f\n",l+1,
	   (genotype->Locus_List+l)->Allele_Name[0],      *(f_allele+i),      *(table->f_allele+i),
	   (genotype->Locus_List+l)->Allele_Name[1],1.0 - *(f_allele+i),1.0 - *(table->f_allele+i));
  }

  return 0;
}

int            print_haplotype_frequency(Estimate_Set  *table,
					 double        *f_noLD,
					 Genotype_Data *genotype)
{
  /* Local variables */
  haplotype   h;
  double      f_total, f_noLD_total, D_total, D;

  /* Print headings */
  printf("\nEstimated haplotype frequency : Loci ");
  for (h=0; h<table->select_loci; h++) {
    printf(" %d",*(table->loci + h)+1);
  }
  printf("\n");

  /* Print frequency */
  f_total      = 0.0;
  f_noLD_total = 0.0;
  D_total      = 0.0;
  printf("     No       p  p_noLD       D   sum p  haplotype\n");
  for (h=0; h<table->nonzero; h++) {
    D = *(table->f + *(table->order + h)) - *(f_noLD   + *(table->order + h));

    f_total      += *(table->f + *(table->order + h));
    f_noLD_total += *(f_noLD   + *(table->order + h));
    D_total      += D;

    printf("%7d %7.4f %7.4f %7.4f %7.4f  ",h+1,
	   *(table->f + *(table->order + h)),*(f_noLD   + *(table->order + h)),D,f_total);
    print_haplotype(*(table->haplo + *(table->order + h)), table, genotype);
    printf("\n");
  }
  printf("  Total %7.4f %7.4f %7.4f\n",f_total,f_noLD_total,D_total);

  return 0;
}

int            print_haplotype_frequency_with_parameters(Estimate_Set  *table,
							 Estimate_Set  *table_for_0,
							 double        *f_noLD,
							 Genotype_Data *genotype)
{
  /* Local variables */
  haplotype   h;
  double      f_total, f_0_total, f_noLD_total, D_total, D;

  /* Print headings */
  printf("\nEstimated haplotype frequency : Loci ");
  for (h=0; h<table->select_loci; h++) {
    printf(" %d",*(table->loci + h)+1);
  }
  printf("\n");

  /* Print frequency */
  f_total      = 0.0;
  f_0_total    = 0.0;
  f_noLD_total = 0.0;
  D_total      = 0.0;
  printf("     No       p     p_0  p_noLD       D   sum p  haplotype\n");
  for (h=0; h<table->nonzero; h++) {
    D = *(table->f + *(table->order + h)) - *(f_noLD   + *(table->order + h));

    f_total      += *(table->f       + *(table->order + h));
    f_0_total    += *(table_for_0->f + *(table->order + h));
    f_noLD_total += *(f_noLD         + *(table->order + h));
    D_total      += D;

    printf("%7d %7.4f %7.4f %7.4f %7.4f %7.4f  ",h+1,
	   *(table->f + *(table->order + h)),
	   *(table_for_0->f + *(table->order + h)),
	   *(f_noLD   + *(table->order + h)),D,f_total);
    print_haplotype(*(table->haplo + *(table->order + h)), table, genotype);
    printf("\n");
  }
  printf("  Total %7.4f %7.4f %7.4f %7.4f\n",f_total,f_0_total,f_noLD_total,D_total);

  return 0;
}

int            print_penetrances(Estimate_Set     *table,
				 double            q_bound[][2],
				 double            confidence,
				 int               conf_flag)
{
  /* Local variables */
  double    conf;

  /* Print penetrances */
  printf("\nEstimated penetrance\n");
  if (conf_flag) {
    conf = confidence * 100.0;
    if (table->model_type == GENOTYPE_MODEL) {
      printf("  Penetrance for no target haplotype  = %6.4f  %3.0f%% CI: %6.4f - %6.4f\n",
	     table->p_plus[0],conf,q_bound[0][0],q_bound[0][1]);
      printf("  Penetrance for 1 target haplotype   = %6.4f  %3.0f%% CI: %6.4f - %6.4f\n",
	     table->p_plus[1],conf,q_bound[1][0],q_bound[1][1]);
      printf("  Penetrance for 2 target haplotypes  = %6.4f  %3.0f%% CI: %6.4f - %6.4f\n",
	     table->p_plus[2],conf,q_bound[2][0],q_bound[2][1]);
    }
    else {
      printf("  Penetrance for target haplotype     = %6.4f  %3.0f%% CI: %6.4f - %6.4f\n",
	     table->p_plus[1],conf,q_bound[1][0],q_bound[1][1]);
      printf("  Penetrance for the other haplotypes = %6.4f  %3.0f%% CI: %6.4f - %6.4f\n",
	     table->p_plus[0],conf,q_bound[0][0],q_bound[0][1]);
    }
  }
  else {
    if (table->model_type == GENOTYPE_MODEL) {
      printf("  Penetrance for no target haplotype  = %6.4f\n",table->p_plus[0]);
      printf("  Penetrance for 1 target haplotype   = %6.4f\n",table->p_plus[1]);
      printf("  Penetrance for 2 target haplotypes  = %6.4f\n",table->p_plus[2]);
    }
    else {
      printf("  Penetrance for target haplotype     = %6.4f\n",table->p_plus[1]);
      printf("  Penetrance for the other haplotypes = %6.4f\n",table->p_plus[0]);
    }
  }

  return 0;
}

int            print_QTL_parameters(Estimate_Set  *table,
				    double            q_bound[][2],
				    double            confidence,
				    int               conf_flag)
{
  /* Local variables */
  double conf;

  /* Print QTL parameters */
  printf("\nEstimated QTL parameters \n");
  if (conf_flag) {
    conf = confidence * 100.0;
    if (table->model_type == GENOTYPE_MODEL) {
      printf("  QTL mean value for the other haplotypes         = %7.4f  %3.0f%% CI: %7.4f - %7.4f\n",
	     table->mu[0],conf,q_bound[0][0],q_bound[0][1]);
      /*  printf("  QTL mean value with 1 target haplotype          = %7.4f  %3.0f%% CI: %7.4f - %7.4f\n",
	  table->mu[1],conf,q_bound[1][0],q_bound[1][1]);*/
      printf("  QTL mean value for the target haplotype         = %7.4f  %3.0f%% CI: %7.4f - %7.4f\n",
	     table->mu[2],conf,q_bound[2][0],q_bound[2][1]);
      printf("  QTL standard deviation                          = %7.4f  %3.0f%% CI: %7.4f - %7.4f\n",
	     table->SD,conf,q_bound[3][0],q_bound[3][1]);
    }
    else {
      printf("  QTL mean value for the target haplotype         = %7.4f  %3.0f%% CI: %7.4f - %7.4f\n",
	     table->mu[1],conf,q_bound[1][0],q_bound[1][1]);
      printf("  QTL mean value for the other haplotypes         = %7.4f  %3.0f%% CI: %7.4f - %7.4f\n",
	     table->mu[0],conf,q_bound[0][0],q_bound[0][1]);
      printf("  QTL standard deviation                          = %7.4f  %3.0f%% CI: %7.4f - %7.4f\n",
	     table->SD,conf,q_bound[2][0],q_bound[2][1]);
    }
  }
  else {
    if (table->model_type == GENOTYPE_MODEL) {
      printf("  QTL mean value for the other haplotypes         = %7.4f\n",table->mu[0]);
      /* printf("  QTL mean value with 1 target haplotype          = %7.4f\n",table->mu[1]);*/
      printf("  QTL mean value for the target haplotype         = %7.4f\n",table->mu[2]);
    }
    else {
      printf("  QTL mean value for the other haplotypes         = %7.4f\n",table->mu[0]);
      printf("  QTL mean value for the target haplotype         = %7.4f\n",table->mu[1]);
    }
    printf("  QTL standard deviation                          = %7.4f\n",table->SD);
  }
  if (table->model_type == GENOTYPE_MODEL) {
    printf("  QTL standard deviation for the other haplotypes = %7.4f\n",table->SD_c[0]);
    /* printf("  QTL standard deviation with 1 target haplotype  = %7.4f\n",table->SD_c[1]);*/
    printf("  QTL standard deviation for the target haplotype = %7.4f\n",table->SD_c[2]);
  }
  else {
    printf("  QTL standard deviation for the other haplotypes = %7.4f\n",table->SD_c[0]);
    printf("  QTL standard deviation for the target haplotype = %7.4f\n",table->SD_c[1]);
  }

  return 0;
}

int            print_haplotype_frequency_by_bootstrap(Estimate_Set  *table,
						      double        *f_0,
						      double        *SE,
						      Genotype_Data *genotype)
{
  /* Local variables */
  haplotype   h;
  double      f_total, f_0_total;

  /* Print headings */
  printf("\nEstimated haplotype frequency by non-parametric bootstrap : Loci ");
  for (h=0; h<table->select_loci; h++) {
    printf(" %d",*(table->loci + h)+1);
  }
  printf("\n");

  /* Print frequency */
  f_total   = 0.0;
  f_0_total = 0.0;
  printf("     No       p  p_boot      SE sum p_boot  haplotype\n");
  for (h=0; h<table->nonzero; h++) {
    f_total   += *(table->f + *(table->order + h));
    f_0_total += *(f_0      + *(table->order + h));

    printf("%7d %7.4f %7.4f %7.4f    %7.4f  ",h+1,
	   *(f_0 + *(table->order + h)),*(table->f + *(table->order + h)),*(SE + *(table->order + h)),f_total);
    print_haplotype(*(table->haplo + *(table->order + h)), table, genotype);
    printf("\n");
  }
  printf("  Total %7.4f %7.4f\n",f_0_total,f_total);

  return 0;
}

int            print_diplotype_profile(Estimate_Set  *table,
				       Genotype_Data *genotype)
{
  /* Local variables */
  int            i, j, s;
  /*  int            same;*/
  haplotype      h;

  /* Probability array */
  double        *p_combi, f;
  haplotype     *order;

  /* Allocate probablity */
  p_combi = (double *)mem_alloc(table->max_haplo_combi*sizeof(double),"Array for combination probability");
  order   = (haplotype *)mem_alloc(table->max_haplo_combi*sizeof(haplotype),"Array fort sorting combination probability");

  /* Print headings */
  printf("\nEstimated diplotype configuration probability without phenotype : Loci ");
  for (i=0; i<table->select_loci; i++) {
    printf(" %d",*(table->loci + i)+1);
  }
  printf("\n");

  /* Print posterier probability */
  printf("     No       name  phenotype  No probability  pop. freq.  diplotype configuration\n");
  for (i=0; i<genotype->Number_of_Subjects; i++) {
    s     = (genotype->Subjects + i)->packed_offset;
    if (s >= 0) {
      /*
      same  = -1;
      if (i > 0) {
	for (j=0; j<i; j++) {
	  if (s == (genotype->Subjects + j)->packed_offset) {
	    same = j+1; break;
	  }
	}
      }
      */

      f = diplotype_probability((table->geno_info + s)->number_of_combi, (table->geno_info + s)->combi, table->f, p_combi);
      hpsort_reverse_index_of_double_array((table->geno_info + s)->number_of_combi, p_combi, order);

      for (j=0; j<(table->geno_info + s)->number_of_combi; j++) {
	if (*(p_combi + *(order + j)) < MIN_FREQUENCY) {
	  break;
	}

	if (j == 0) {
	  printf("%7d %10s %10d %3d     %7.4f     %7.4f  ",
		 i+1,(genotype->Subjects + i)->Name,(genotype->Subjects + i)->phenotype,
		 j+1,*(p_combi + *(order+j)),*(p_combi + *(order+j))*f);
	}
	else {
	  printf("                              %3d     %7.4f     %7.4f  ",
		 j+1,*(p_combi + *(order+j)),*(p_combi + *(order+j))*f);
	}

	h = ((table->geno_info + s)->combi + *(order + j))->diplotype[0];
	print_haplotype(*(table->haplo + h), table, genotype);
	printf(" ");
	h = ((table->geno_info + s)->combi + *(order + j))->diplotype[1];
	print_haplotype(*(table->haplo + h), table, genotype);

	/*
	if (j == 0) {
	  if (same >= 0) {
	    printf(" same as subject %d",same);
	  }
	}
	*/

	printf("\n");
      }
    }
  }

  return 0;
}

int            print_diplotype_profile_with_penetrance(Estimate_Set  *table,
						       Genotype_Data *genotype)
{
  /* Local variables */
  int            i, j, s;
  /*  int            same;*/
  haplotype      h;

  /* Probability array */
  double        *p_combi, f;
  haplotype     *order;

  /* Allocate probablity */
  p_combi = (double *)mem_alloc(table->max_haplo_combi*sizeof(double),"Array for combination probability");
  order   = (haplotype *)mem_alloc(table->max_haplo_combi*sizeof(haplotype),"Array fort sorting combination probability");

  /* Print headings */
  printf("\nEstimated incidence rate for each diplotype configuration with phenotype : Loci ");
  for (i=0; i<table->select_loci; i++) {
    printf(" %d",*(table->loci + i)+1);
  }
  printf("\n");

  /* Print posterier probability */
  printf("     No       name  phenotype  No probability  pop. freq.  diplotype configuration\n");
  for (i=0; i<genotype->Number_of_Subjects; i++) {
    s     = (genotype->Subjects + i)->packed_offset;
    if (s >= 0) {
      /*
      same  = -1;
      if (i > 0) {
	for (j=0; j<i; j++) {
	  if (s == (genotype->Subjects + j)->packed_offset) {
	    same = j+1; break;
	  }
	}
      }
      */

      f = diplotype_probability_with_penetrance((table->geno_info + s)->number_of_combi, (table->geno_info + s)->combi, (table->geno_info + s)->class, (table->geno_info + s)->phenotype, table, table->f, p_combi);
      hpsort_reverse_index_of_double_array((table->geno_info + s)->number_of_combi, p_combi, order);

      for (j=0; j<(table->geno_info + s)->number_of_combi; j++) {
	if (*(p_combi + *(order + j)) < MIN_FREQUENCY) {
	  break;
	}

	if (j == 0) {
	  printf("%7d %10s %10d %3d     %7.4f     %7.4f  ",
		 i+1,(genotype->Subjects + i)->Name,(genotype->Subjects + i)->phenotype,
		 j+1,*(p_combi + *(order+j)),*(p_combi + *(order+j))*f);
	}
	else {
	  printf("                              %3d     %7.4f     %7.4f  ",
		 j+1,*(p_combi + *(order+j)),*(p_combi + *(order+j))*f);
	}

	h = ((table->geno_info + s)->combi + *(order + j))->diplotype[0];
	print_haplotype(*(table->haplo + h), table, genotype);
	printf(" ");
	h = ((table->geno_info + s)->combi + *(order + j))->diplotype[1];
	print_haplotype(*(table->haplo + h), table, genotype);

	/*
	if (j == 0) {
	  if (same >= 0) {
	    printf(" same as subject %d",same);
	  }
	}
	*/

	printf("\n");
      }
    }
  }

  return 0;
}

int            print_diplotype_profile_with_QTL(Estimate_Set  *table,
						Genotype_Data *genotype)
{
  /* Local variables */
  int            i, j, s, same;
  haplotype      h;

  /* Probability array */
  double        *p_combi;
  haplotype     *order;

  /* Allocate probablity */
  p_combi = (double *)mem_alloc(table->max_haplo_combi*sizeof(double),"Array for combination probability");
  order   = (haplotype *)mem_alloc(table->max_haplo_combi*sizeof(haplotype),"Array fort sorting combination probability");

  /* Print headings */
  printf("\nEstimated incidence rate for each diplotype configuration with QTL : Loci ");
  for (i=0; i<table->select_loci; i++) {
    printf(" %d",*(table->loci + i)+1);
  }
  printf("\n");

  /* Print posterier probability */
  printf("     No       name  phenotype  No probability  diplotype configuration\n");
  for (i=0; i<genotype->Number_of_Subjects; i++) {
    s     = (genotype->Subjects + i)->packed_offset;
    if (s >= 0) {
      same  = -1;
      if (i > 0) {
	for (j=0; j<i; j++) {
	  if (s == (genotype->Subjects + j)->packed_offset) {
	    same = j+1; break;
	  }
	}
      }

      diplotype_probability_with_QTL((table->geno_info + s)->number_of_combi, (table->geno_info + s)->combi, (table->geno_info + s)->class, (genotype->Subjects + i)->QTL, table, table->f, p_combi);
      hpsort_reverse_index_of_double_array((table->geno_info + s)->number_of_combi, p_combi, order);

      for (j=0; j<(table->geno_info + s)->number_of_combi; j++) {
	if (*(p_combi + *(order + j)) < MIN_FREQUENCY) {
	  break;
	}

	if (j == 0) {
	  printf("%7d %10s %10.3e %3d     %7.4f  ",
		 i+1,(genotype->Subjects + i)->Name,(genotype->Subjects + i)->QTL,
		 j+1,*(p_combi + *(order+j)));
	}
	else {
	  printf("                              %3d     %7.4f  ",
		 j+1,*(p_combi + *(order+j)));
	}

	h = ((table->geno_info + s)->combi + *(order + j))->diplotype[0];
	print_haplotype(*(table->haplo + h), table, genotype);
	printf(" ");
	h = ((table->geno_info + s)->combi + *(order + j))->diplotype[1];
	print_haplotype(*(table->haplo + h), table, genotype);

	if (j == 0) {
	  if (same >= 0) {
	    printf(" same as subject %d",same);
	  }
	}

	printf("\n");
      }
    }
  }

  return 0;
}

int            print_test_of_haplotype_set_with_penetrance(Estimate_Set  *table,
							   Genotype_Data *genotype,
							   Haplotype_Set *set_of_haplotype,
							   unsigned int  *order,
							   double         confidence,
							   int            conf_flag)
{
  /* Local variables */
  int               i, s;
  double            risk_ratio, conf;
  Haplotype_Subset *subset;

  /* Print headings */
  printf("\nResult of association test with haplotype set\n");
  if (table->model_type == GENOTYPE_MODEL) {
    printf("  No.  statistics     P value      q0      q1      q2  haplotype set\n");
  }
  else {
    printf("  No.  statistics     P value      q-      q+      OR* haplotype set\n");
  }

  subset = set_of_haplotype->subset;
  for (i=0; i<set_of_haplotype->number_of_subset; i++) {
    s = *(order + i);
    if (table->model_type == GENOTYPE_MODEL) {
      printf("%5d %11.4f %11.4e %7.4f %7.4f %7.4f  ",
	     s+1,(subset+s)->statistic,(subset+s)->p_value,
	     (subset+s)->result[0],(subset+s)->result[1],(subset+s)->result[2]);
    }
    else {
      printf("%5d %11.4f %11.4e %7.4f %7.4f %s  ",
	     s+1,(subset+s)->statistic,(subset+s)->p_value,
	     (subset+s)->result[0],(subset+s)->result[1],sprint_odds_ratio_7((subset+s)->odds_ratio));
    }
    print_haplotype_with_mask((subset + s)->target, (subset + s)->mask, table, genotype);
    printf("\n");
  }

  if (conf_flag) {
    conf = confidence * 100.0;
    printf("\n %3.0f%% confidence interval of estimates\n",conf);
    if (table->model_type == GENOTYPE_MODEL) {
      printf("  No.      CI of q0         CI of q1         CI of q2     haplotype set\n");
    }
    else {
      printf("  No.      CI of q-         CI of q+          CI of OR*      haplotype set\n");
    }
    for (i=0; i<set_of_haplotype->number_of_subset; i++) {
      s = *(order + i);
      if (table->model_type == GENOTYPE_MODEL) {
	printf("%5d  %6.4f - %6.4f  %6.4f - %6.4f  %6.4f - %6.4f  ",s+1,
	       (subset+s)->interval[0][0],(subset+s)->interval[0][1],
	       (subset+s)->interval[1][0],(subset+s)->interval[1][1],
	       (subset+s)->interval[2][0],(subset+s)->interval[2][1]);
      }
      else {
	printf("%5d  %6.4f - %6.4f  %6.4f - %6.4f %s - ",s+1,
	       (subset+s)->interval[0][0],(subset+s)->interval[0][1],
	       (subset+s)->interval[1][0],(subset+s)->interval[1][1],
	       sprint_odds_ratio((subset+s)->odds_interval[0]));

	printf("%s  ",sprint_odds_ratio((subset+s)->odds_interval[1]));
      }
      print_haplotype_with_mask((subset + s)->target, (subset + s)->mask, table, genotype);
      printf("\n");
    }
  }

  if (table->model_type != GENOTYPE_MODEL) {
    printf("\n  * Odds ratio for case-control study\n");
  }

  return 0;
}

int            print_test_of_haplotype_set_with_QTL(Estimate_Set  *table,
						    Genotype_Data *genotype,
						    Haplotype_Set *set_of_haplotype,
						    unsigned int  *order,
						    double         confidence,
						    int            conf_flag)
{
  /* Local variables */
  int               i, s;
  double            risk_ratio, conf;
  Haplotype_Subset *subset;

  /* Print headings */
  printf("\nResult of association test with haplotype set\n");
  if (table->model_type == GENOTYPE_MODEL) {
    printf("  No.  statistics     P value         mu0         mu1         mu2          SD  haplotype set\n");
  }
  else {
    printf("  No.  statistics     P value         mu-         mu+          SD  haplotype set\n");
  }

  subset = set_of_haplotype->subset;
  for (i=0; i<set_of_haplotype->number_of_subset; i++) {
    s = *(order + i);
    if (table->model_type == GENOTYPE_MODEL) {
      printf("%5d %11.4f %11.4e %11.4e %11.4e %11.4e %11.4e  ",
	     s+1,(subset+s)->statistic,(subset+s)->p_value,
	     (subset+s)->result[0],(subset+s)->result[1],(subset+s)->result[2],(subset+s)->result[3]);
    }
    else {
      printf("%5d %11.4f %11.4e %11.4e %11.4e %11.4e  ",
	     s+1,(subset+s)->statistic,(subset+s)->p_value,
	     (subset+s)->result[0],(subset+s)->result[1],(subset+s)->result[3]);
    }
    print_haplotype_with_mask((subset + s)->target, (subset + s)->mask, table, genotype);
    printf("\n");
  }

  if (conf_flag) {
    conf = confidence * 100.0;
    printf("\n %3.0f%% confidence interval of estimates\n",conf);
    if (table->model_type == GENOTYPE_MODEL) {
      printf("  No.           CI of mu0                  CI of mu1                  CI of mu2                  CI of SD          haplotype set\n");
    }
    else {
      printf("  No.           CI of mu-                  CI of mu+                  CI of SD          haplotype set\n");
    }
    for (i=0; i<set_of_haplotype->number_of_subset; i++) {
      s = *(order + i);
      if (table->model_type == GENOTYPE_MODEL) {
	printf("%5d  %11.4e - %11.4e  %11.4e - %11.4e  %11.4e - %11.4e  %11.4e - %11.4e  ",s+1,
	       (subset+s)->interval[0][0],(subset+s)->interval[0][1],
	       (subset+s)->interval[1][0],(subset+s)->interval[1][1],
	       (subset+s)->interval[2][0],(subset+s)->interval[2][1],
	       (subset+s)->interval[3][0],(subset+s)->interval[3][1]);
      }
      else {
	printf("%5d  %11.4e - %11.4e  %11.4e - %11.4e  %11.4e - %11.4e  ",s+1,
	       (subset+s)->interval[0][0],(subset+s)->interval[0][1],
	       (subset+s)->interval[1][0],(subset+s)->interval[1][1],
	       (subset+s)->interval[2][0],(subset+s)->interval[2][1]);
      }
      print_haplotype_with_mask((subset + s)->target, (subset + s)->mask, table, genotype);
      printf("\n");
    }
  }

  return 0;
}

int            print_haplotype(haplotype      haplo,
			       Estimate_Set  *table,
			       Genotype_Data *genotype)
{
  fprint_haplotype(stdout, haplo, table, genotype);

  return 0;
}

int            fprint_haplotype(FILE          *fp,
				haplotype      haplo,
				Estimate_Set  *table,
				Genotype_Data *genotype)
{
  int l, a;

  for (l=0; l<table->select_loci; l++) {
    a = (haplo >> (table->select_loci-l-1)) & 01;
    fprintf(fp,"%s",(genotype->Locus_List + *(table->loci + l))->Allele_Name[a]);
  }

  return 0;
}

int            print_haplotype_with_mask(haplotype      haplo,
					 haplotype      mask,
					 Estimate_Set  *table,
					 Genotype_Data *genotype)
{
  fprint_haplotype_with_mask(stdout, haplo, mask, table, genotype);

  return 0;
}

int            fprint_haplotype_with_mask(FILE          *fp,
					  haplotype      haplo,
					  haplotype      mask,
					  Estimate_Set  *table,
					  Genotype_Data *genotype)
{
  int l, a, m;

  for (l=0; l<table->select_loci; l++) {
    a = (haplo >> (table->select_loci-l-1)) & 01;
    m = (mask  >> (table->select_loci-l-1)) & 01;
    if (m == 1) {
      fprintf(fp,"%s",(genotype->Locus_List + *(table->loci + l))->Allele_Name[a]);
    }
    else {
      fprintf(fp,"%s",MISSING_DATA_NAME);
    }
  }

  return 0;
}

int            sprint_haplotype_with_mask(char          *haplotype_name,
					  haplotype      haplo,
					  haplotype      mask,
					  Estimate_Set  *table,
					  Genotype_Data *genotype)
{
  int l, a, m;

  *haplotype_name = '\0';
  for (l=0; l<table->select_loci; l++) {
    a = (haplo >> (table->select_loci-l-1)) & 01;
    m = (mask  >> (table->select_loci-l-1)) & 01;
    if (m == 1) {
      strcat(haplotype_name,(genotype->Locus_List + *(table->loci + l))->Allele_Name[a]);
    }
    else {
      strcat(haplotype_name,MISSING_DATA_NAME);
    }
  }

  return 0;
}

char          *sprint_odds_ratio(double odds_ratio)
{
  static char  buff[12];

  if (odds_ratio < 0.0) {
    strcpy(buff, "      NA");
  }
  else {
    if (odds_ratio < 1.0e3) {
      sprintf(buff,"%8.4f",odds_ratio);
    }
    else if (odds_ratio < 1.0e4) {
      sprintf(buff,"%8.3f",odds_ratio);
    }
    else {
      sprintf(buff,"%8.1e",odds_ratio);
    }
  }

  return buff;
}

char          *sprint_odds_ratio_7(double odds_ratio)
{
  static char  buff[8];

  if (odds_ratio < 0.0) {
    strcpy(buff, "     NA");
  }
  else {
    if (odds_ratio < 1.0e2) {
      sprintf(buff,"%7.4f",odds_ratio);
    }
    else if (odds_ratio < 1.0e3) {
      sprintf(buff,"%7.3f",odds_ratio);
    }
    else if (odds_ratio < 1.0e4) {
      sprintf(buff,"%7.2f",odds_ratio);
    }
    else {
      sprintf(buff,"%7.0e",odds_ratio);
    }
  }

  return buff;
}
