#include "haplotypetest.h"

/* Utility function */
extern void          *mem_alloc(size_t, const char *);
extern void          *mem_realloc(void *, size_t, const char *);
extern void           hpsort_index_of_int_array(int, unsigned int *, unsigned int *);

/* Error handler */
extern int            error_check();
extern int            error_message(char *);
extern int            warning_message(char *);

#ifdef DEBUG_COMBI
extern char          *binary_string(haplotype);
#endif

/* Define functions */
Estimate_Set  *combination(Estimate_Set *);
int            set_genotype(Genotype_Pack *, int, Diplotype *, int [][3], int *, haplotype *, haplotype *);
int            all_haplotype(Estimate_Set *);
int            set_diplotype_class(Estimate_Set *);

Estimate_Set  *combination(Estimate_Set *table)
{
  /* Local variables */
  int             s, l, number_of_hetero, count;
  haplotype       c, number_of_combination;

  /* Variables to make Combination */
  Diplotype       control[MAX_NUMBER_OF_LOCI];
  int             missing[MAX_NUMBER_OF_LOCI][3];
  haplotype       Genotype_Code, Hetero_Mask;
  haplotype       hetero[MAX_NUMBER_OF_LOCI];
  haplotype       hetero_mask;

  /* Variables for combination storage */
  int             size_of_diplo, last_of_diplo;
  Diplotype      *diplo;

#ifdef DEBUG_COMBI
  int             debug_set = 2;
#endif

  /* Set locus bit to control array */
  for (l=0; l<table->select_loci; l++) {
    control[l].diplotype[0] = 00;
    control[l].diplotype[1] = 01 << l;
  }

  /* Make combination table */
  size_of_diplo = 0;
  for (s=0; s<table->number_of_set; s++) {
    (table->geno_info + s)->number_of_combi = 0;

#ifdef DEBUG_COMBI
      printf("# Set No. %5d : phenotype = %d, weight = %f\n",s,(table->geno_info + s)->phenotype,(table->geno_info + s)->weight);
      printf("#      miss  %32s",binary_string((table->geno_info + s)->Missing[0]));
      printf(" %32s\n",binary_string((table->geno_info + s)->Missing[1]));
#endif

    /* Set genotype including missing data */
    while (set_genotype(table->geno_info + s, table->select_loci, control, missing, &count, &Genotype_Code, &Hetero_Mask)) {
    
      if (Hetero_Mask == 0) { /* Homozygous */
	/* Get number of combination */
	number_of_combination = 1;

	/* Allocate combination table */
	last_of_diplo  = size_of_diplo;
	size_of_diplo += number_of_combination;
	table->combi   = (Diplotype *)mem_realloc(table->combi,size_of_diplo*sizeof(Diplotype),"Haplotype combination table");
	diplo          = table->combi + last_of_diplo;
	(table->geno_info + s)->number_of_combi += number_of_combination;

	/* Make combination */
	diplo->diplotype[0] = Genotype_Code;
	diplo->diplotype[1] = Genotype_Code;
	diplo++;
      }
      else {                  /* Heterozygous */
	/* Get number of heterozygous loci and the position */
	number_of_hetero = 0;
	for (l=0; l<table->select_loci; l++) {
	  if (control[l].diplotype[1] & Hetero_Mask) {
	    hetero[number_of_hetero] = l; number_of_hetero++;
	  }
	}

	/* Get number of combination */
	number_of_hetero--;
	number_of_combination = 01 << number_of_hetero;

	/* Allocate combination table */
	last_of_diplo  = size_of_diplo;
	size_of_diplo += number_of_combination;
	table->combi   = (Diplotype *)mem_realloc(table->combi,size_of_diplo*sizeof(Diplotype),"Haplotype combination table");
	diplo          = table->combi + last_of_diplo;
	(table->geno_info + s)->number_of_combi += number_of_combination;

	/* Make combination */
	for (c=0; c<number_of_combination; c++) {
	  hetero_mask = 0;
	  for (l=0; l<number_of_hetero; l++) {
	    hetero_mask = hetero_mask | control[hetero[l]].diplotype[(c >> l) & 01];
	  }
	  diplo->diplotype[0] = hetero_mask | Genotype_Code;
	  diplo->diplotype[1] = diplo->diplotype[0] ^ Hetero_Mask;
	  diplo++;
	}
      }
      /* End of one genotype */
    }
    /* End of one subject */
  }

  /* Set last pointer information */
  table->max_haplo_combi = 0;
  table->total_combi     = 0;
  diplo                  = table->combi;
  for (s=0; s<table->number_of_set; s++) {
    (table->geno_info + s)->combi = diplo;
    diplo = diplo + (table->geno_info + s)->number_of_combi;
    table->total_combi += (table->geno_info + s)->number_of_combi;
    if (table->max_haplo_combi < (table->geno_info + s)->number_of_combi) {
      table->max_haplo_combi = (table->geno_info + s)->number_of_combi;
    }
  }
#ifdef DEBUG_COMBI
  printf("# Maximum number of combination = %d\n",table->max_haplo_combi);
  for (s=0; s<table->number_of_set; s++) {
    printf("# Set No. %5d : combination = %5d",s,(table->geno_info + s)->number_of_combi);
    printf(" %8.8x\n",(table->geno_info + s)->combi);
    if (s == debug_set) {
      printf("#               : code1 = %32s\n",binary_string((table->geno_info + s)->Genotype_Code[0]));
      printf("#               : code2 = %32s\n",binary_string((table->geno_info + s)->Genotype_Code[1]));
      printf("#               : miss1 = %32s\n",binary_string((table->geno_info + s)->Missing[0]));
      printf("#               : miss2 = %32s\n",binary_string((table->geno_info + s)->Missing[1]));
      diplo = (table->geno_info + s)->combi;
      for (c=0; c<(table->geno_info + s)->number_of_combi; c++) {
	printf("#            %32s",binary_string(diplo->diplotype[0]));
	printf(" %32s\n",binary_string(diplo->diplotype[1]));
	diplo++;
      }
    }
  }
#endif

  return table;
}

int            set_genotype(Genotype_Pack *info,
			    int            select_loci,
			    Diplotype     *control,
			    int            missing[][3],
			    int           *count,
			    haplotype     *Genotype_Code,
			    haplotype     *Hetero_Mask)
{
  /* Local variables */
  int       RC, l;
  haplotype genotype[2];
  haplotype mask[2];

  RC          = 1;
  genotype[0] = info->Genotype_Code[0];
  genotype[1] = info->Genotype_Code[1];
  if (info->number_of_combi == 0) {
    /* Initialize flags */
    *count = 0;
    if (info->Missing[0]) { /* There are some missing data */
      for (l=0; l<select_loci; l++) {
	if (info->Missing[0] & control[l].diplotype[1]) {
	  missing[*count][0] = l;
	  missing[*count][1] = 1;
	  missing[*count][2] = 0;
	  if (info->Missing[1] & control[l].diplotype[1]) {
	    missing[*count][1] = 2;
	  }
	  *count += 1;
	}
      }
    }
  }
  else {
    /* Next genotype */
    if (*count > 0) {
      /* Make missing genotype pattern */
      missing[0][2] += 1;
      for (l=0; l<*count; l++) {
	if (missing[l][2] > missing[l][1]) {
	  if (l+1 < *count) {
	    missing[l+1][2] += 1;
	    missing[l  ][2]  = 0;
	  }
	  else {
	    RC = 0; /* Done */
	  }
	}
      }
      if (RC) {
	/* Set missing genotype pattern */
	mask[0] = 00;
	mask[1] = 00;
	for (l=0; l<*count; l++) {
	  if (missing[l][2] != 0) {
	    mask[1] = mask[1] | control[missing[l][0]].diplotype[1];
	  }
	  if (missing[l][2] == 2) {
	    mask[0] = mask[0] | control[missing[l][0]].diplotype[1];
	  }	
	}
	/* Final genotype */
#ifdef DEBUG_COMBI
	printf("#     mask : %32s",binary_string(mask[0]));
	printf(" %32s\n",binary_string(mask[1]));
#endif
	genotype[0] = genotype[0] | mask[0];
	genotype[1] = genotype[1] | mask[1];
      }
    }
    else {
      RC = 0;
    }
  }

  *Genotype_Code = genotype[0];
  *Hetero_Mask   = genotype[0] ^ genotype[1];

#ifdef DEBUG_COMBI
  if (RC) {
    printf("# genotype : %32s",binary_string(genotype[0]));
    printf(" %32s\n",binary_string(genotype[1]));
  }
#endif

  return RC;
}

int            all_haplotype(Estimate_Set *table)
{
  /* Local variables */
  int             s, number_of_haplotype;
  haplotype      *combi, *index, h, h_index;
  unsigned char  *h_hlag[2];
#ifdef DEBUG_COMBI
  Diplotype      *diplo;
  haplotype       c;
#endif

  number_of_haplotype = table->total_combi * 2;

  combi = (haplotype *)table->combi;
  table->haplo = (haplotype *)mem_alloc(number_of_haplotype*sizeof(haplotype),"Possible haplotype array");
  index = table->haplo;

  hpsort_index_of_int_array(number_of_haplotype, combi, index);

  h_index = 0;
  *index = *(combi + *(index));
  for (h=1; h<number_of_haplotype; h++) {
    if (*(combi + *(index+h)) == *(index+h_index)) {
      *(combi + *(index+h)) = h_index;
    }
    else {
      h_index++;
      *(index+h_index) = *(combi + *(index+h));
      *(combi + *(index+h)) = h_index;
    }
  }

  /* Release unused ares */
  table->number_of_haplo = h_index+1;
  table->haplo = (haplotype *)mem_realloc(table->haplo, table->number_of_haplo*sizeof(haplotype), "Possible haplotype array");

#ifdef DEBUG_COMBI
  for (h=0; h<table->number_of_haplo; h++) {
    printf("# haplotype %32s",binary_string(*(table->haplo + h)));
    printf(" %32s\n",binary_string(table->target));
  }
  for (s=0; s<table->number_of_set; s++) {
    diplo = (table->geno_info + s)->combi;

    for (c=0; c<(table->geno_info + s)->number_of_combi; c++) {
      printf("# %5d %5d\n",diplo->diplotype[0],diplo->diplotype[1]);
      diplo++;
    }
  }
#endif

  return 0;
}

int            set_diplotype_class(Estimate_Set *table)
{
  /* Local variables */
  int             s, c;
  Diplotype      *combi;
  DiploClass     *class;

  /* Allocate diplotype class table */
  if (table->class == NULL) {
    table->class  = (DiploClass *)mem_alloc(table->total_combi*sizeof(DiploClass), "Diplotype class array for possible haplotype");
  }

  /* Loop for genotype set */
  class         = table->class;
  for (s=0; s<table->number_of_set; s++) {
    (table->geno_info + s)->class = class;
    combi = (table->geno_info + s)->combi;
    for (c=0; c<(table->geno_info + s)->number_of_combi; c++) {
      if ((*(table->haplo + combi->diplotype[0]) & table->target_mask) == table->target) {
	class->h_flag[0] = (unsigned char)1;
      }
      else {
	class->h_flag[0] = (unsigned char)0;
      }
      if ((*(table->haplo + combi->diplotype[1]) & table->target_mask) == table->target) {
	class->h_flag[1] = (unsigned char)1;
      }
      else {
	class->h_flag[1] = (unsigned char)0;
      }
      combi++; class++;
    }
  }

  return 0;
}
