#include "haplotypetest.h"

/* Utility function */
extern void          *mem_alloc(size_t, const char *);
extern void          *mem_realloc(void *, size_t, const char *);
extern void           hpsort_reverse_index_of_double_array(int, double *, unsigned int *);
extern void           sort_int_array(int, unsigned int *); 

/* Error handler */
extern int            error_check();
extern int            error_message(char *);
extern int            warning_message(char *);

/* Result print */
extern int            print_haplotype_with_mask(haplotype, haplotype, Estimate_Set *, Genotype_Data *);
extern int            print_haplotype(haplotype, Estimate_Set *, Genotype_Data *);

#ifdef DEBUG_HAPLOTYPE_SET
extern char          *binary_string(haplotype);
#endif

/* Define functions */
Haplotype_Set *set_haplotype_set(Genotype_Data *, Estimate_Set *, Option_Data *);
int            get_htSNP_loci(Genotype_Data *, Estimate_Set *, Haplotype_Set *, Option_Data *);
int            set_htSNP_loci(Estimate_Set *, Haplotype_Set *, Option_Data *);
int            omit_redundant_locus(Estimate_Set *, allele**, char [], int, int);
int            find_set_of_haplotype_set(Genotype_Data *, Estimate_Set *, Haplotype_Set *, Option_Data *);
int            count_up_haplotype_set(int, int, char *);
int            register_haplotype_set(Haplotype_Set *, Haplotype_Subset **, int *);
int            register_haplotype_set_with_patterncheck(Haplotype_Set *, Haplotype_Subset **, int *);
Haplotype_Set *alloc_haplotype_set();
void           free_haplotype_set(Haplotype_Set *);

Haplotype_Set *set_haplotype_set(Genotype_Data *genotype,
				 Estimate_Set  *table,
				 Option_Data   *option)
{
  /* Local variables */
  Haplotype_Set  *set_of_haplotype;
  double          f_total;
  haplotype       h, member;
  int             s, first;
  char            space[34];

  set_of_haplotype = alloc_haplotype_set();

  if (option->select_htSNP == 0) {
    get_htSNP_loci(genotype, table, set_of_haplotype, option);
  }
  else {
    set_htSNP_loci(table, set_of_haplotype, option);
  }

  if (option->msg_flag) {
    printf("\nHaplotype represented by htSNP\n");
    printf("     No       p   sum p  haplotype\n");
    f_total = 0.0;
    for (h=0; h<table->nonzero; h++) {
      f_total      += *(table->f + *(table->order + h));
      printf("%7d %7.4f %7.4f  ",h+1,*(table->f + *(table->order + h)),f_total);
      print_haplotype(*(table->haplo + *(table->order + h)), table, genotype);
      printf(" => ");
      print_haplotype_with_mask(*(table->haplo + *(table->order + h)), set_of_haplotype->loci_mask, table, genotype);
      printf("\n");
    }
  }

  if (option->haplotype_set) {
    find_set_of_haplotype_set(genotype, table, set_of_haplotype, option);

    if (option->msg_flag) {
      printf("\nInformation of SNP anchored haplotype set\n  No.  haplotype set representation     member of haplotype set\n");
      strcpy(space,"");
      for (s=0; s<33 - table->select_loci; s++) {
	strcat(space," ");
      }
      for (s=0; s<set_of_haplotype->number_of_subset; s++) {
	printf("%5d  ",s+1);
	print_haplotype_with_mask((set_of_haplotype->subset + s)->target, (set_of_haplotype->subset + s)->mask, table, genotype);
	printf("%s",space);
	member = (set_of_haplotype->subset + s)->member;
	first  = 1;
	for (h=0; h<table->nonzero; h++) {
	  if (member & 01) {
	    if (first) {
	      first = 0;
	    }
	    else {
	      printf("                                        ");
	    }
	    print_haplotype_with_mask(*(table->haplo + h), set_of_haplotype->loci_mask, table, genotype);
	    printf(" (");
	    print_haplotype(*(table->haplo + h), table, genotype);
	    printf(")\n");
	  }
	  member = member >> 1;
	}
      }
    }
  }

  return set_of_haplotype;
}

int            get_htSNP_loci(Genotype_Data *genotype,
			      Estimate_Set  *table,
			      Haplotype_Set *set_of_haplotype,
			      Option_Data   *option)
{
  /* Local variables */
  allele          **SNP; /* [table->nonzero][table->select_loci]; */ /* 11/09/06 */ 
  char            locus_pattern[table->select_loci][2][table->nonzero+1];
  int             number_of_htSNP, index_array[table->select_loci];
  haplotype       h, haplo;
  int             l, m, l_max;
  char            site_info[table->select_loci];
  char            site_index[table->select_loci], site_index_work[table->select_loci];
  allele          a, not_a;
  double          min_major_allele[table->select_loci];

  /* added on 11/09/06 */
  SNP = (allele **) malloc (table->nonzero * sizeof (allele*));
  for(h=0; h<table->nonzero; h++) {
    SNP[h] = (allele *) malloc(table->select_loci * sizeof(allele));
  }

  /* SNP table */
  for (h=0; h<table->nonzero; h++) {
    haplo = *(table->haplo + h);
    for (l=0; l<table->select_loci; l++) {
      SNP[h][l]           = (haplo >> (table->select_loci-l-1)) & 01;
    }
  }

  /* Locus pattern table */
  for (l=0; l<table->select_loci; l++) {
    for (h=0; h<table->nonzero; h++) {
      locus_pattern[l][0][h] = 060 + SNP[h][l];
      locus_pattern[l][1][h] = 061 - SNP[h][l];
    }
    locus_pattern[l][0][table->nonzero] = STRING_TERMINATOR;
    locus_pattern[l][1][table->nonzero] = STRING_TERMINATOR;
  }

#ifdef DEBUG_HAPLOTYPE_SET
  for (h=0; h<table->nonzero; h++) {
    printf("# haplotype %2d ",h+1);
    for (l=0; l<table->select_loci; l++) {
      printf("%d",SNP[h][l]);
    }
    printf("\n");
  }
  for (l=0; l<table->select_loci; l++) {
    printf("# %3d %s %s\n",l+1,locus_pattern[l][0],locus_pattern[l][1]);
  }
#endif

  /* Get informative locus */
  for (l=0; l<table->select_loci; l++) {
    site_info[l] = -99;
  }

  for (l=0; l<table->select_loci; l++) {
    a = SNP[0][l];
    for (h=1; h<table->nonzero; h++) {
      if (SNP[h][l] != a) {
	site_info[l] = -1; break;
      }
    }
  }

  for (l=0; l<table->select_loci-1; l++) {
    if (site_info[l] == -1) {
      for (m=l+1; m<table->select_loci; m++) {
	if (site_info[m] == -1) {
	  if (strcmp(locus_pattern[l][0],locus_pattern[m][0]) == 0 ||
	      strcmp(locus_pattern[l][0],locus_pattern[m][1]) == 0) {
	    site_info[m] = l;
	  }
	}
      }
    }
  }

#ifdef DEBUG_HAPLOTYPE_SET
  for (l=0; l<table->select_loci; l++) {
    printf("#     locus %2d ",l+1);
    for (h=0; h<table->nonzero; h++) {
      printf("%d",SNP[h][l]);
    }
    printf(" %3d\n",site_info[l]);
  }
#endif

  /* Count informative locus */
  number_of_htSNP = 0;
  for (l=0; l<table->select_loci; l++) {
    if (site_info[l] == -1) {
      site_index_work[number_of_htSNP] = l;
      number_of_htSNP++;
    }
  }

  /* Find the maximum of minor allele frequency in same group */
  for (l=0; l<number_of_htSNP; l++) {
    min_major_allele[l] = (genotype->Locus_List + *(table->loci + site_index_work[l]))->Allele_Frequency;
#ifdef DEBUG_HAPLOTYPE_SET
    printf("# at locus %3d : major allele frequncy = %f\n",site_index_work[l],(genotype->Locus_List + *(table->loci + site_index_work[l]))->Allele_Frequency);
#endif
    l_max = site_index_work[l];
    for (m=0; m<table->select_loci; m++) {
      if (site_info[m] == site_index_work[l]) {
	if (min_major_allele[l] > (genotype->Locus_List + *(table->loci + m))->Allele_Frequency) {
	  min_major_allele[l] = (genotype->Locus_List + *(table->loci + m))->Allele_Frequency;
	  l_max = m;
	}
#ifdef DEBUG_HAPLOTYPE_SET
	printf("# at locus %3d : major allele frequncy = %f : %3d\n",m,(genotype->Locus_List + *(table->loci + m))->Allele_Frequency,l_max);
#endif
      }
    }

    if (site_index_work[l] != l_max) {
      for (m=0; m<table->select_loci; m++) {
	if (site_info[m] == site_index_work[l]) {
	  site_info[m]              = l_max;
	}
      }
      site_info[l_max]              = -1;
      site_info[site_index_work[l]] = l_max;
      site_index_work[l]            = l_max;
    }
  }

  if (option->msg_flag) {
    printf("\nhtSNP defined by haplotype  \n");
    printf("  locus  locus pattern\n");
    for (l=0; l<table->select_loci; l++) {
      printf("%7d  ",*(table->loci+l) + 1);
      for (h=0; h<table->nonzero; h++) {
	printf("%s",(genotype->Locus_List + *(table->loci + l))->Allele_Name[(int)SNP[h][l]]);
      }
      if      (site_info[l] == -99) {
	printf(" : not informative\n");
      }
      else if (site_info[l] == -1) {
	printf(" : informative\n");
      }
      else if (site_info[l] >=  0) {
	printf(" : same as locus %d\n",*(table->loci+site_info[l])+1);
      }
    }
  }

  hpsort_reverse_index_of_double_array(number_of_htSNP, min_major_allele, index_array);

  for (l=0; l<number_of_htSNP; l++) {
    site_index[l] = site_index_work[index_array[l]];
  }  

#ifdef DEBUG_HAPLOTYPE_SET
  for (l=0; l<table->select_loci; l++) {
    printf("#     locus %3d ",l+1);
    for (h=0; h<table->nonzero; h++) {
      printf("%d",SNP[h][l]);
    }
    printf(" %3d\n",site_info[l]);
  }

  for (l=0; l<table->select_loci; l++) {
    if (site_info[l] == -1) {
      printf("#     locus %3d ",l+1);
      for (h=0; h<table->nonzero; h++) {
	printf("%d",SNP[h][l]);
      }
      printf(" %3d\n",site_info[l]);
    }
  }

  printf("# number of htSNP = %d\n",number_of_htSNP);
  for (l=0; l<number_of_htSNP; l++) {
    printf("# %3d %f\n",site_index_work[l],min_major_allele[l]);
  }
  printf("# number of htSNP = %d\n",number_of_htSNP);
  for (l=0; l<number_of_htSNP; l++) {
    printf("# %3d %f\n",site_index[l],min_major_allele[index_array[l]]);
  }
#endif

  number_of_htSNP = omit_redundant_locus(table,SNP,site_index,number_of_htSNP,0);

  set_of_haplotype->number_of_htSNP = number_of_htSNP;
  set_of_haplotype->htSNP_loci      = (int *)mem_alloc(number_of_htSNP*sizeof(int),"Table for htSNP");
  for (l=0; l<number_of_htSNP; l++) {
    *(set_of_haplotype->htSNP_loci + l) = site_index[l];
  }
  sort_int_array(number_of_htSNP, set_of_haplotype->htSNP_loci);
  set_of_haplotype->loci_mask       = 0;
  for (l=0, m=0; l<table->select_loci; l++) {
    if (l == *(set_of_haplotype->htSNP_loci + m)) {
      set_of_haplotype->loci_mask = (set_of_haplotype->loci_mask << 1) | 01;
      m++;
    }
    else {
      set_of_haplotype->loci_mask = (set_of_haplotype->loci_mask << 1);
    }
  }
#ifdef DEBUG
  printf("# loci mask : %32s\n",binary_string(set_of_haplotype->loci_mask));
#endif

  if (option->msg_flag) {
    printf("\nFinal htSNP loci :");
    for (l=0; l<set_of_haplotype->number_of_htSNP; l++) {
      printf(" %d",*(table->loci + *(set_of_haplotype->htSNP_loci + l)) + 1);
    }
    printf("\n");
  }


  /* added on 11/09/06 */
  for(h=0; h<table->nonzero; h++) free (SNP[h]);
  free (SNP);
  
  return 0;
}

int            set_htSNP_loci(Estimate_Set  *table,
			      Haplotype_Set *set_of_haplotype,
			      Option_Data   *option)
{
  /* Local variables */
  int    l, s, number_of_htSNP;
  int    htSNP[table->select_loci];

  number_of_htSNP = 0;
  for (l=0; l<table->select_loci; l++) {
    for (s=0; s<option->select_htSNP; s++) {
      if (*(table->loci + l) == *(option->loci_htSNP+s)-1) {
	htSNP[number_of_htSNP] = l;
	number_of_htSNP++;
      }
    }
  }

  set_of_haplotype->number_of_htSNP = number_of_htSNP;
  set_of_haplotype->htSNP_loci      = (int *)mem_alloc(number_of_htSNP*sizeof(int),"Table for htSNP");
  for (l=0; l<number_of_htSNP; l++) {
    *(set_of_haplotype->htSNP_loci + l) = htSNP[l];
  }
  sort_int_array(number_of_htSNP, set_of_haplotype->htSNP_loci);
  set_of_haplotype->loci_mask       = 0;
  for (l=0, s=0; l<table->select_loci; l++) {
    if (l == *(set_of_haplotype->htSNP_loci + s)) {
      set_of_haplotype->loci_mask = (set_of_haplotype->loci_mask << 1) | 01;
      s++;
    }
    else {
      set_of_haplotype->loci_mask = (set_of_haplotype->loci_mask << 1);
    }
  }

  if (option->msg_flag) {
    printf("\nSpecified htSNP loci :");
    for (l=0; l<set_of_haplotype->number_of_htSNP; l++) {
      printf(" %d",*(table->loci + *(set_of_haplotype->htSNP_loci + l)) + 1);
    }
    printf("\n");
  }

  return 0;
}


int            omit_redundant_locus(Estimate_Set  *table,
				    allele         **SNP, /* [][table->select_loci], */ /* 11/09/06 */
				    char           site_index[],
				    int            number_of_htSNP,
				    int            start_SNP)
{
  /* Local variables */ 
  haplotype   haplo[table->nonzero], h, i;
  int         RC;
  int         l, s, sh, differ, tmp_htSNP, min_htSNP, f_omit;
  char        site_index_s[number_of_htSNP], site_index_w[number_of_htSNP-1];

  RC        = number_of_htSNP;
  min_htSNP = number_of_htSNP;
#ifdef DEBUG_HAPLOTYPE_SET
  printf("# omit redundant locus\n");
  printf("# number of htSNP = %d :",number_of_htSNP);
  for (l=0; l<number_of_htSNP; l++) {
    printf(" %d",site_index[l]);
  }
  printf("\n");
#endif

  f_omit = 0;
  for (l=start_SNP; l<number_of_htSNP; l++) {
    /* Make haplotype with out aimed locus */
    for (h=0; h<table->nonzero; h++) {
      haplo[h] = 0;
      for (s=0, sh=0; s<number_of_htSNP; s++) {
	if (s != l) {
	  haplo[h] = haplo[h] | (SNP[h][site_index[s]] << sh);
	  sh++;
	}
      }
#ifdef DEBUG_HAPLOTYPE_SET
      printf("# %d %32s\n",h,binary_string(haplo[h]));
#endif
    }

    differ = 1;
    for (h=0; h<table->nonzero-1; h++) {
      for (i=h+1; i<table->nonzero; i++) {
	if (haplo[h] == haplo[i]) {
	  differ = 0;
	}
      }
    }

    if (differ) {
#ifdef DEBUG_HAPLOTYPE_SET
      printf("# locus %d (order %d) can be reduced\n",site_index[l],l);
#endif      
      for (s=0; s<=l; s++) {
	site_index_w[s] = site_index[s];
      }
      for (s=l+1; s<number_of_htSNP; s++) {
	site_index_w[s-1] = site_index[s];
      }
      tmp_htSNP = number_of_htSNP - 1;
      tmp_htSNP = omit_redundant_locus(table,SNP,site_index_w,tmp_htSNP,l);
      if (tmp_htSNP < min_htSNP) {
	f_omit = 1;
	min_htSNP = tmp_htSNP;
	for (s=0; s<min_htSNP; s++) {
	  site_index_s[s] = site_index_w[s];
	}
      }
    }
  }

  if (f_omit) {
    for (l=0; l<min_htSNP; l++) {
      site_index[l] = site_index_s[l];
    }
  }
#ifdef DEBUG_HAPLOTYPE_SET
  printf("# number of htSNP = %d :",min_htSNP);
  for (l=0; l<min_htSNP; l++) {
    printf(" %d",site_index[l]);
  }
  printf("\n");
#endif

  return min_htSNP;
}

int            find_set_of_haplotype_set(Genotype_Data *genotype,
					 Estimate_Set  *table,
					 Haplotype_Set *set_of_haplotype,
					 Option_Data   *option)
{
  /* Local variables */
  char              count[set_of_haplotype->number_of_htSNP];
  int               total_set, buff_size, l, c, r, s;
  Haplotype_Subset *subset;
  haplotype         h, all_mask;
  haplotype loci[set_of_haplotype->number_of_htSNP], mask[set_of_haplotype->number_of_htSNP];

  /* Allocate table */
  buff_size = TEMP_BLOCK_SIZE;
  set_of_haplotype->subset = (Haplotype_Subset *)mem_alloc(buff_size*sizeof(Haplotype_Subset),"Table for haplotype subset");
  set_of_haplotype->number_of_subset = 0;
  subset                   = set_of_haplotype->subset;

  /* First stage : complete haplotypes */
  all_mask = 0;
  for (l=0; l<table->select_loci; l++) {
    all_mask = (all_mask << 1) | 01;
  }
  for (h=0; h<table->nonzero; h++) {
#ifdef DEBUG_HAPLOTYPE_SET
    printf("# %s\n",binary_string(*(table->haplo + h)));
#endif
    subset->target = *(table->haplo + h);
    subset->mask   = all_mask;
    subset->member = 01 << h;

    register_haplotype_set(set_of_haplotype, &subset, &buff_size);
  }

  /* Second stage : complete haplotypes represented by htSNP */
  total_set = 1;
  for (l=0; l<set_of_haplotype->number_of_htSNP; l++) {
    count[l]  = 0;
    total_set = total_set*2;
    loci[l]   =  (01 << (table->select_loci - *(set_of_haplotype->htSNP_loci+l)-1));
    mask[l]   = ~(01 << (table->select_loci - *(set_of_haplotype->htSNP_loci+l)-1));
#ifdef DEBUG_HAPLOTYPE_SET
    printf("# %s\n",binary_string(loci[l]));
    printf("  %s\n",binary_string(mask[l]));
#endif
  }

#ifdef DEBUG_HAPLOTYPE_SET
  printf("# Expected number of haplotype set = %d\n",total_set);
#endif

  /* Get complete haplotypes */
  for (c=0; c<total_set; c++) {
#ifdef DEBUG_HAPLOTYPE_SET
    printf("# %5d : ",c+1);
    for (l=0; l<set_of_haplotype->number_of_htSNP; l++) {
      printf("%d",count[l]);
    }
    printf("\n");
#endif

    subset->target = 0;
    subset->mask   = set_of_haplotype->loci_mask;
    subset->member = 0;
    for (l=0; l<set_of_haplotype->number_of_htSNP; l++) {
      if      (count[l] == 1) {
	subset->target = subset->target | loci[l];
      }
    }

    for (h=0; h<table->nonzero; h++) {
#ifdef DEBUG_HAPLOTYPE_SET
      printf("# %s\n",binary_string(*(table->haplo + h)));
#endif
      if ((*(table->haplo + h) & subset->mask) == subset->target) {
	subset->member = subset->member | (01 << h);
#ifdef DEBUG_HAPLOTYPE_SET
	printf("  %s\n",binary_string(subset->mask));
	printf("  %s\n",binary_string(subset->target));
#endif
      }
    }

    if (subset->member != 0) {
      register_haplotype_set_with_patterncheck(set_of_haplotype, &subset, &buff_size);
    }

    count_up_haplotype_set(set_of_haplotype->number_of_htSNP, 2, count);
  }

  /* Third stage : SNPs */
  for (s=0; s<set_of_haplotype->number_of_htSNP; s++) {
    for (c=0; c<set_of_haplotype->number_of_htSNP; c++) {
      count[c] = 2;
    }

    for (c=0; c<2; c++) {
      count[s]--;
#ifdef DEBUG_HAPLOTYPE_SET
      printf("# %5d : ",c+1);
      for (l=0; l<set_of_haplotype->number_of_htSNP; l++) {
	printf("%d",count[l]);
      }
      printf("\n");
#endif

      subset->target = 0;
      subset->mask   = set_of_haplotype->loci_mask;
      subset->member = 0;
      for (l=0; l<set_of_haplotype->number_of_htSNP; l++) {
	if      (count[l] == 1) {
	  subset->target = subset->target | loci[l];
	}
	else if (count[l] == 2) {
	  subset->mask   = subset->mask   & mask[l];
	}
      }

      for (h=0; h<table->nonzero; h++) {
#ifdef DEBUG_HAPLOTYPE_SET
	printf("# %s\n",binary_string(*(table->haplo + h)));
#endif
	if ((*(table->haplo + h) & subset->mask) == subset->target) {
	  subset->member = subset->member | (01 << h);
#ifdef DEBUG_HAPLOTYPE_SET
	  printf("  %s\n",binary_string(subset->mask));
	  printf("  %s\n",binary_string(subset->target));
#endif
	}
      }

      if (subset->member != 0) {
	register_haplotype_set(set_of_haplotype, &subset, &buff_size);
      }
    }
  }

  /* Forth stage : incomplete haplotypes */
  total_set = 1;
  for (l=0; l<set_of_haplotype->number_of_htSNP; l++) {
    count[l]  = 0;
    total_set = total_set*3;
  }
  count[0] = 2;
  total_set = total_set-3;
#ifdef DEBUG_HAPLOTYPE_SET
  printf("# Expected number of haplotype set = %d\n",total_set);
#endif

  /* Get complete haplotypes */
  for (c=0; c<total_set; c++) {
#ifdef DEBUG_HAPLOTYPE_SET
    printf("# %5d : ",c+1);
    for (l=0; l<set_of_haplotype->number_of_htSNP; l++) {
      printf("%d",count[l]);
    }
    printf("\n");
#endif

    subset->target = 0;
    subset->mask   = set_of_haplotype->loci_mask;
    subset->member = 0;
    for (l=0; l<set_of_haplotype->number_of_htSNP; l++) {
      if      (count[l] == 1) {
	subset->target = subset->target | loci[l];
      }
      else if (count[l] == 2) {
	subset->mask   = subset->mask   & mask[l];
      }
    }

    for (h=0; h<table->nonzero; h++) {
#ifdef DEBUG_HAPLOTYPE_SET
      printf("# %s\n",binary_string(*(table->haplo + h)));
#endif
      if ((*(table->haplo + h) & subset->mask) == subset->target) {
	subset->member = subset->member | (01 << h);
#ifdef DEBUG_HAPLOTYPE_SET
	printf("  %s\n",binary_string(subset->mask));
	printf("  %s\n",binary_string(subset->target));
#endif
      }
    }

    if (subset->member != 0) {
      register_haplotype_set(set_of_haplotype, &subset, &buff_size);
    }

    count_up_haplotype_set(set_of_haplotype->number_of_htSNP, 3, count);
  }

#ifdef DEBUG_HAPLOTYPE_SET
  printf("# Expected number of haplotype set = %d\n",total_set);
#endif

  set_of_haplotype->subset = (Haplotype_Subset *)mem_realloc(set_of_haplotype->subset, set_of_haplotype->number_of_subset*sizeof(Haplotype_Subset),"Table for haplotype subset");
#ifdef DEBUG
  printf("# number of haplotype subset = %d\n",set_of_haplotype->number_of_subset);
  subset     = set_of_haplotype->subset;
  for (c=0; c<set_of_haplotype->number_of_subset; c++) {
    printf("# target : %s\n",binary_string(subset->target));
    printf("  mask   : %s\n",binary_string(subset->mask));
    printf("  member : %s\n",binary_string(subset->member));
    subset++;
  }
#endif

  return 0;
}

int            count_up_haplotype_set(int  number_of_loci,
				      int  base,
				      char count[])
{
  /* Local variables */
  int    RC = 1, l;

  count[0]++;

  for (l=0; l<number_of_loci-1; l++) {
    if (count[l] == base) {
      count[l] = 0;
      count[l+1]++;
    }
  }

  return RC;
}

int            register_haplotype_set(Haplotype_Set     *set_of_haplotype,
				      Haplotype_Subset **subset,
				      int               *buff_size)
{
  /* Local variables */
  int   r, l;

  r = 1;
  for (l=0; l<set_of_haplotype->number_of_subset; l++) {
    if ((*subset)->member == (set_of_haplotype->subset + l)->member) {
      r = 0; break;
    }
  }
  if (r) {
    set_of_haplotype->number_of_subset += 1;
    (*subset)++;
    if (set_of_haplotype->number_of_subset == *buff_size) {
      *buff_size += TEMP_BLOCK_SIZE;
      l          = *subset - set_of_haplotype->subset;
      set_of_haplotype->subset = (Haplotype_Subset *)mem_realloc(set_of_haplotype->subset, *buff_size*sizeof(Haplotype_Subset),"Table for haplotype subset");
      *subset    = set_of_haplotype->subset + l;
    }
  }

  return r;
}

int            register_haplotype_set_with_patterncheck(Haplotype_Set     *set_of_haplotype,
						   Haplotype_Subset **subset,
						   int               *buff_size)
{
  /* Local variables */
  int   r, l;

  r = 1;
  for (l=0; l<set_of_haplotype->number_of_subset; l++) {
    if ((*subset)->target == (set_of_haplotype->subset + l)->target &&	
	(*subset)->mask   == (set_of_haplotype->subset + l)->mask) {
      r = 0; break;
    }
  }
  if (r) {
    set_of_haplotype->number_of_subset += 1;
    (*subset)++;
    if (set_of_haplotype->number_of_subset == *buff_size) {
      *buff_size += TEMP_BLOCK_SIZE;
      l          = *subset - set_of_haplotype->subset;
      set_of_haplotype->subset = (Haplotype_Subset *)mem_realloc(set_of_haplotype->subset, *buff_size*sizeof(Haplotype_Subset),"Table for haplotype subset");
      *subset    = set_of_haplotype->subset + l;
    }
  }

  return r;
}

Haplotype_Set *alloc_haplotype_set()
{
  /* Local variables */
  Haplotype_Set *set_of_haplotype;
  
  set_of_haplotype = (Haplotype_Set *)mem_alloc(sizeof(Haplotype_Set),"Table for haplotype set");
  set_of_haplotype->htSNP_loci = NULL;
  set_of_haplotype->subset     = NULL;

  return set_of_haplotype;
}

void           free_haplotype_set(Haplotype_Set *set_of_haplotype)
{
  if (set_of_haplotype->htSNP_loci != NULL) {
    free(set_of_haplotype->htSNP_loci);
  }
  if (set_of_haplotype->subset     != NULL) {
    free(set_of_haplotype->subset);
  }
  free(set_of_haplotype);

  return;
}

