#include "haplotypetest.h"

/* Utility function */
extern void          *mem_alloc(size_t, const char *);
extern void           hpsort_index_of_double_array(int, double *, unsigned int *);

/* Error handler */
extern int            error_check();
extern int            error_message(char *);
extern int            warning_message(char *);

/* Define functions */
int            get_phenotype(Genotype_Data *, Option_Data *, char *);
int            get_QTLvalue(Genotype_Data *, Option_Data *, char *);
int            logistic(int, int [], double []);
int            blank_check(char *);
int            get_phenotype_with_ID(char *, char *, char **);
int            get_QTL_ID(char *, char *, double *);

int            get_phenotype(Genotype_Data *genotype,
			     Option_Data   *option,
			     char          *phenotype_file)
{
  int            subj_count, i, current, RC, not_found;
  int            value;
  char          *real_buffer, *ptr, *str;
  char           ID[CHAR_IN_SUBJECT_NAME+1];


  /* Open phenotype file */
  if (strlen(phenotype_file) == 0) {
    return 0;
  }
  else {
    File_Status->phenotype_fp = fopen(phenotype_file,"r");
  }

  if (File_Status->phenotype_fp == NULL) {
    error_message("phenotype file open error.");
    if (error_check()) return 1;
  }

  for(i=0; i<genotype->Number_of_Subjects; i++){
    (genotype->Subjects + i)->phenotype    = -1;
  }

  /* Copy phenotype file name */
  strcpy(File_Status->phenotype_name, phenotype_file);

  real_buffer = (char *)mem_alloc(MAX_BUFFER_LENGTH*sizeof(char),"Read buffer");
  subj_count  = 0;
  while(fgets(real_buffer, MAX_BUFFER_LENGTH-1, File_Status->phenotype_fp) != NULL) {
    /* Check record length */
    ptr = strchr(real_buffer, CARRIAGE_RETURN);
    if (ptr == NULL) {
      ptr = strchr(real_buffer, LINE_FEED);
    }
    if (ptr == NULL) {
      error_message("phenotype file input buffer over flow. Change MAX_RECORD_LENGTH.");
      if (error_check()) {
	free(real_buffer); return 1;
      }
    }

    /* Delete \n */
    *ptr = STRING_TERMINATOR;

    if (option->phenotype_format) {
      not_found = 1;
      ptr       = real_buffer;
      while (*ptr != STRING_TERMINATOR) {
	if ((*ptr != DELIMITER_SPACE) && (*ptr != DELIMITER_TAB)) {
	  if (not_found) {
	    not_found = 0;
	    str = ptr;
	  }
	}
	else {
	  *ptr = STRING_TERMINATOR;
	}
	ptr++;
      }
      if (!not_found) {
	if      (strcmp(str,"0") == 0) {
	  if (option->switch_p) {
	    (genotype->Subjects + subj_count)->phenotype = 1;
	  }
	  else {
	    (genotype->Subjects + subj_count)->phenotype = 0;
	  }
	}
	else if (strcmp(str,"1") == 0) {
	  if (option->switch_p) {
	    (genotype->Subjects + subj_count)->phenotype = 0;
	  }
	  else {
	    (genotype->Subjects + subj_count)->phenotype = 1;
	  }
	}
      }
      subj_count++;
      if (subj_count >= genotype->Number_of_Subjects) {
	break;
      }
    }
    else { /* with_ID */
      RC = get_phenotype_with_ID(real_buffer, ID, &str);
      if      (RC == 2) {
	if      (strcmp(str,"0") == 0) {
	  if (option->switch_p) {
	    value = 1;
	  }
	  else {
	    value = 0;
	  }
	}
	else if (strcmp(str,"1") == 0) {
	  if (option->switch_p) {
	    value = 0;
	  }
	  else {
	    value = 1;
	  }
	}
	not_found = 1;
	for(i=0; i<genotype->Number_of_Subjects; i++){
	  if (strcmp((genotype->Subjects + i)->Name, ID) == 0) {
	    (genotype->Subjects + i)->phenotype    = value;
	    not_found = 0;
	    break;
	  }
	}
	if (not_found) {
	  printf("  Subject ID %s having phenotype %s in phenotype file is not found in genotype file.\n",ID,str);
	}
      }
      else if (RC == 1) {
	not_found = 1;
	for(i=0; i<genotype->Number_of_Subjects; i++){
	  if (strcmp((genotype->Subjects + i)->Name, ID) == 0) {
	    (genotype->Subjects + i)->phenotype    = -1;
	    break;
	  }
	}
	if (not_found) {
	  printf("  Subject ID %s having no phenotype data in phenotype file is not found in genotype file.\n",ID);
	}
      }
    }
  }
  if (option->check_ID) {
    for(i=0; i<genotype->Number_of_Subjects; i++){
      printf("  Subject ID %s",(genotype->Subjects + i)->Name);
      if ((genotype->Subjects + i)->phenotype == -1) {
	printf(" has no phenotype.\n");
      }
      else {
	printf(" has phenotype value %d.\n",(genotype->Subjects + i)->phenotype);
      }
    }
  }

#ifdef DEBUG
  for (subj_count=0; subj_count <genotype->Number_of_Subjects; subj_count++) {
    printf("# %5d : %2d\n",subj_count,
	   (genotype->Subjects + subj_count)->phenotype);
  }
#endif

  free(real_buffer);

  return 0;
}

int            get_QTLvalue(Genotype_Data *genotype,
			    Option_Data   *option,
			    char          *QTL_file)
{
  int            subj_count, i, current, RC, not_found;
  double         value;
  int           *order;
  unsigned int  *index;
  double        *QTL_value;
  char          *real_buffer, *ptr;
  char           ID[CHAR_IN_SUBJECT_NAME+1];

  /* Open QTL file */
  if (strlen(QTL_file) == 0) {
    error_message("QTL file is not specified.");
    if (error_check()) return 1;
  }
  else {
    File_Status->phenotype_fp = fopen(QTL_file,"r");
  }

  if (File_Status->phenotype_fp == NULL) {
    error_message("QTL file open error.");
    if (error_check()) return 1;
  }

  for(i=0; i<genotype->Number_of_Subjects; i++){
    (genotype->Subjects + i)->phenotype    = -1;
  }

  /* Copy QTL file name */
  strcpy(File_Status->phenotype_name, QTL_file);

  real_buffer = (char *)mem_alloc(MAX_BUFFER_LENGTH*sizeof(char),"Read buffer");
  subj_count  = 0;
  while(fgets(real_buffer, MAX_BUFFER_LENGTH-1, File_Status->phenotype_fp) != NULL) {
    /* Check record length */
    ptr = strchr(real_buffer, LINE_FEED);
    if (ptr == NULL) {
      error_message("QTL phenotype file input buffer over flow. Change MAX_RECORD_LENGTH.");
      if (error_check()) {
	free(real_buffer); return 1;
      }
    }

    /* Delete \n */
    *ptr = STRING_TERMINATOR;

    if (option->phenotype_format) {
      if (blank_check(real_buffer)) {
	sscanf(real_buffer,"%lf",&value);
	(genotype->Subjects + subj_count)->phenotype    = 1;
	(genotype->Subjects + subj_count)->QTL_value[0] = value;
      }
      else {
	(genotype->Subjects + subj_count)->QTL_value[0] = 0;
      }
#ifdef DEBUG
      printf("# %5d : %2d : %f\n",subj_count,
	     (genotype->Subjects + subj_count)->phenotype,
	     (genotype->Subjects + subj_count)->QTL_value[0]);
#endif
      subj_count++;
      if (subj_count >= genotype->Number_of_Subjects) {
	break;
      }
    }
    else { /* with_ID */
      RC = get_QTL_ID(real_buffer, ID, &value);
      if      (RC == 2) {
	not_found = 1;
	for(i=0; i<genotype->Number_of_Subjects; i++){
	  if (strcmp((genotype->Subjects + i)->Name, ID) == 0) {
	    (genotype->Subjects + i)->phenotype    = 1;
	    (genotype->Subjects + i)->QTL_value[0] = value;
	    not_found = 0;
	    break;
	  }
	}
	if (not_found) {
	  printf("  Subject ID %s having phenotype %g in phenotype file is not found in genotype file.\n",ID,value);
	}
      }
      else if (RC == 1) {
	not_found = 1;
	for(i=0; i<genotype->Number_of_Subjects; i++){
	  if (strcmp((genotype->Subjects + i)->Name, ID) == 0) {
	    (genotype->Subjects + i)->phenotype    = -1;
	    not_found = 0;
	    break;
	  }
	}
	if (not_found) {
	  printf("  Subject ID %s having no phenotype data in phenotype file is not found in genotype file.\n",ID);
	}
      }
    }
  }
  if (option->check_ID) {
    for(i=0; i<genotype->Number_of_Subjects; i++){
      printf("  Subject ID %s",(genotype->Subjects + i)->Name);
      if ((genotype->Subjects + i)->phenotype == -1) {
	printf(" has no phenotype.\n");
      }
      else {
	printf(" has phenotype value %f.\n",(genotype->Subjects + i)->QTL_value[0]);
      }
    }
  }
  free(real_buffer);

  fclose(File_Status->phenotype_fp);
  File_Status->phenotype_fp = NULL;

  if (option->convertQTL) {
    index     = (unsigned int *)mem_alloc(genotype->Number_of_Subjects*sizeof(unsigned int),"Array for sort QTL phenotype");
    order     = (int          *)mem_alloc(genotype->Number_of_Subjects*sizeof(int),         "Array for sort QTL phenotype");
    QTL_value = (double       *)mem_alloc(genotype->Number_of_Subjects*sizeof(double),      "Array for sort QTL phenotype");

    subj_count = 0;
    for(i=0; i<genotype->Number_of_Subjects; i++){
      if ((genotype->Subjects + i)->phenotype != -1) {
	*(QTL_value + subj_count) = (genotype->Subjects + i)->QTL_value[0];
	subj_count++;
      }
    }

    hpsort_index_of_double_array(subj_count, QTL_value, index);

#ifdef DEBUG
    for(i=0; i<subj_count; i++){
      printf("# %5d %13.5e %5d\n",i,*(QTL_value + i),*(index + i));
    }
#endif

    current = 0;
    value   = *(QTL_value + *index);
    for(i=0; i<subj_count; i++){
      if (*(QTL_value + *(index + i)) == value) {
	*(order + i) = current;
      }
      else {
	current = i;
	*(order + i) = current;
	value        = *(QTL_value + *(index + i));
      }
    }
#ifdef DEBUG
    printf("# index value\n");
    for(i=0; i<subj_count; i++){
      printf("# %5d %13.5e %d\n",i,*(QTL_value + *(index + i)), *(order+i));
    }
#endif

    logistic(subj_count, order, QTL_value);

    subj_count = 0;
    for(i=0; i<genotype->Number_of_Subjects; i++){
      if ((genotype->Subjects + i)->phenotype != -1) {
	*(order + subj_count) = i;
	subj_count++;
      }
    }

#ifdef DEBUG
    printf("# index value order\n");
    for(i=0; i<subj_count; i++){
      printf("# %5d %13.5e %5d %5d %5d\n",i,*(QTL_value + i),
	     *(index + i),*(order + i),
	     (genotype->Subjects + *(order + i))->phenotype);
    }
#endif

    subj_count = 0;
    for(i=0; i<genotype->Number_of_Subjects; i++){
      if ((genotype->Subjects + i)->phenotype != -1) {
	(genotype->Subjects + *(order + *(index + subj_count)))->QTL_value[1] = *(QTL_value + subj_count);
	subj_count++;
      }
      else {
	(genotype->Subjects + i)->QTL_value[1] = 0.0;
      }
    }

#ifdef DEBUG
    for(i=0; i<genotype->Number_of_Subjects; i++){
      printf("# %5d %5d %13.5e %13.5e\n",i,(genotype->Subjects + i)->phenotype,
	     (genotype->Subjects + i)->QTL_value[0],
	     (genotype->Subjects + i)->QTL_value[1]);
    }
#endif

    for(i=0; i<genotype->Number_of_Subjects; i++){
      (genotype->Subjects + i)->QTL = (genotype->Subjects + i)->QTL_value[1];
    }

    free(QTL_value);
    free(order);
    free(index);
  }
  else {
    for(i=0; i<genotype->Number_of_Subjects; i++){
      (genotype->Subjects + i)->QTL = (genotype->Subjects + i)->QTL_value[0];
    }
  }

  return 0;
}

int            logistic(int     number_of_data,
			int     order[],
			double  result[])
{
  int    i;
  int    counter1;
  int    counter2;
  double sum, y;

  counter1 = 0;
  do {
    counter2 = 0;

    while (counter1!=0 && counter1<number_of_data && order[counter1+counter2]==order[counter1-1]) {
      ++counter2;
    }

    sum=0;

    for (i=0; i<counter2+1; ++i) {
      y    = ((double)(counter1+i-1)+0.5)/(double)(number_of_data);
      sum += -0.5586 * log((1-y)/y); 
    }

    if (counter1 == 0) {
      result[counter1] = sum/(double)(counter2+1);
    }
    else {
      for (i=0; i<counter2+1; ++i) {
	result[counter1+i-1] = sum/(double)(counter2+1);
      }
    }

    counter1 += counter2+1;
  } while(counter1 < number_of_data+1);

  return(sum);
}

int            blank_check(char *buff)
{
  int   RC = 0;
  char *ptr;

  ptr = strchr(buff, CARRIAGE_RETURN);
  if (ptr == NULL) {
    ptr = strchr(buff, LINE_FEED);
  }
  if (ptr != NULL) {
    *ptr = STRING_TERMINATOR;
  }

  ptr = buff;
  while (*ptr != STRING_TERMINATOR) {
    if ((*ptr != DELIMITER_SPACE) && (*ptr != DELIMITER_TAB)) {
      RC = 1;
      break;
    }
    ptr++;
  }

  return RC;
}

int            get_QTL_ID(char   *buff,
			  char   *ID,
			  double *QTL_value)
{
  int   RC = 0;
  char *ptr, *str[2];
  int   item, count, size;

  /* Check # */
  ptr = strchr(buff, COMMENT_CHAR);
  if      (ptr == buff) {
    return RC;
  }

  /* Check \r */
  if (ptr == NULL) {
    ptr = strchr(buff, CARRIAGE_RETURN);
  }

  /* Get character string */
  if (ptr != NULL) {
    *ptr = STRING_TERMINATOR;
  }

  ptr    = buff;
  str[0] = NULL;
  str[1] = NULL;
  count  = 0;
  item   = 0;
  while (*ptr != STRING_TERMINATOR) {
    if ((*ptr != DELIMITER_SPACE) && (*ptr != DELIMITER_TAB)) {
      if (!item) {
	if (count == 2) {
	  break;
	}
	str[count] = ptr;
	/*	printf("# QTL %x str[%d] %x\n",buff,count,str[count]);*/
	count++; item = 1;
      }
    }
    else {
      if (item) {
	*ptr = STRING_TERMINATOR;
	item = 0;
      }
    }
    ptr++;
  }

  if (str[0] != NULL) {
    if (strlen(str[0]) > CHAR_IN_SUBJECT_NAME) {
      sprintf(error_message_buffer,"Subject name, %s, is too long. First %d characters are used.",str[0],CHAR_IN_SUBJECT_NAME);
      warning_message(error_message_buffer);
      *(str[0]+CHAR_IN_SUBJECT_NAME) = STRING_TERMINATOR;
    }
    strcpy(ID,str[0]);
    /*    printf("|%s|",ID);*/
    RC = 1;
  }
  if (str[1] != NULL) {
    sscanf(str[1],"%lf",QTL_value);
    /*    printf("|%f|",*QTL_value);*/
    RC = 2;
  }

  /*  printf("|%d|\n",RC);*/
  return RC;
}

int            get_phenotype_with_ID(char   *buff,
				     char   *ID,
				     char  **phenotype)
{
  /* Local variables */
  int   RC = 0;
  char *ptr, *str[2];
  int   item, count, size;

  /* Check # */
  ptr = strchr(buff, COMMENT_CHAR);
  if      (ptr == buff) {
    return RC;
  }

  /* Check \r */
  if (ptr == NULL) {
    ptr = strchr(buff, CARRIAGE_RETURN);
  }

  /* Get character string */
  if (ptr != NULL) {
    *ptr = STRING_TERMINATOR;
  }

  ptr    = buff;
  str[0] = NULL;
  str[1] = NULL;
  count  = 0;
  item   = 0;
  while (*ptr != STRING_TERMINATOR) {
    if ((*ptr != DELIMITER_SPACE) && (*ptr != DELIMITER_TAB)) {
      if (!item) {
	if (count == 2) {
	  break;
	}
	str[count] = ptr;
	count++; item = 1;
      }
    }
    else {
      if (item) {
	*ptr = STRING_TERMINATOR;
	item = 0;
      }
    }
    ptr++;
  }

  if (str[0] != NULL) {
    if (strlen(str[0]) > CHAR_IN_SUBJECT_NAME) {
      sprintf(error_message_buffer,"Subject name, %s, is too long. First %d characters are used.",str[0],CHAR_IN_SUBJECT_NAME);
      warning_message(error_message_buffer);
      *(str[0]+CHAR_IN_SUBJECT_NAME) = STRING_TERMINATOR;
    }
    strcpy(ID,str[0]);
    RC = 1;
  }
  if (str[1] != NULL) {
    *phenotype = str[1];
    RC = 2;
  }

  /*  printf("|%d|\n",RC);*/
  return RC;
}
