#include "haplotypetest.h"

/* Utility function */
extern void          *mem_alloc(size_t, const char *);

/* Error handler */
extern int            error_check();
extern int            error_message(char *);
extern int            warning_message(char *);

#ifdef DEBUG
extern char          *binary_string(haplotype);
#endif

/* Define functions */
int            set_target_haplotype(Genotype_Data *, Estimate_Set *, Haplotype_Set *, Option_Data *);

int            set_target_haplotype(Genotype_Data *genotype,
				    Estimate_Set  *table,
				    Haplotype_Set *set_of_haplotype,
				    Option_Data   *option)
{
  char        *tmp, *ptr, *dlm;
  int          loop, locus_count, locus_No, l;
  int          a, not_found;
  Locus_Data  *locus;
  int          RC;

  set_of_haplotype->subset = (Haplotype_Subset *)mem_alloc(sizeof(Haplotype_Subset),"Table for haplotype subset");
  set_of_haplotype->number_of_subset = 1;
  (set_of_haplotype->subset)->target = 0;
  (set_of_haplotype->subset)->mask   = 0;
  (set_of_haplotype->subset)->member = 0;

  l                  = strlen(option->target);
  tmp                = (char *)mem_alloc((l+1)*sizeof(char),"Target haplotype decode area");
  strcpy(tmp,option->target);
  ptr                = tmp;
  locus_count        = 0;

  RC                 = 0;
  loop               = 1;
  while (loop) {
    dlm = ptr + 1;
    while ((*dlm != TARGET_LOCI_DELM) && (*dlm != '\0')) {
      dlm++;
    }
    if (*dlm == '\0') {
      loop = 0;
    }
    *dlm = '\0';

    locus_No = -1;
    for (l=0; l<table->select_loci; l++) {
      if (locus_count == *(table->loci + l)) {
	locus_No = *(table->loci + l);
	break;
      }
    }
    if (locus_No >= 0) {
      locus    = genotype->Locus_List + locus_No;
#ifdef DEBUG
      printf("# locus = %5d : allele = %s %s : target allele = %s\n",
	     locus_No,locus->Allele_Name[0],locus->Allele_Name[1],ptr);
#endif

      not_found = 1;
      for(a=0; a<locus->number_of_allele; a++) {
	if (strcmp(ptr,locus->Allele_Name[a]) == 0) {
	  not_found = 0;
	  (set_of_haplotype->subset)->target = ((set_of_haplotype->subset)->target << 1) |  a;
	  (set_of_haplotype->subset)->mask   = ((set_of_haplotype->subset)->mask   << 1) | 01;
	}
      }
      if (not_found) {
	if (strcmp(ptr,MISSING_DATA_NAME) == 0) {
	  (set_of_haplotype->subset)->target = ((set_of_haplotype->subset)->target << 1);
	  (set_of_haplotype->subset)->mask   = ((set_of_haplotype->subset)->mask   << 1);
	}
	else {
	  sprintf(error_message_buffer,"Allele %s of locus %d of target haplotype is not found in genotype.",ptr,locus_No+1);
	  error_message(error_message_buffer);
	}
      }
    }

    locus_count++;
    ptr = dlm + 1;
  }
  free(tmp);

  RC = error_check();

#ifdef DEBUG
  printf("# target haplotype      = %s\n",binary_string((set_of_haplotype->subset)->target));
  printf("# target haplotype_mask = %s\n",binary_string((set_of_haplotype->subset)->target));
#endif

  return 0;
}
