#include "haplotypetest.h"

/* Utility function */
extern void          *mem_alloc(size_t, const char *);
extern void           hpsort_reverse_index_of_double_array(int, double *, unsigned int *);

/* Error handler */
extern int            error_check();
extern int            error_message(char *);
extern int            warning_message(char *);

/* Define functions */
int            trait_inheritance_model_Q1(Estimate_Set *, int, int);
static int            no_association_model(Estimate_Set *);
static int            dominant_model(Estimate_Set *);
static int            recessive_model(Estimate_Set *);
static int            genotype_model(Estimate_Set *);
static int            over_dominant_model(Estimate_Set *);
static int            print_inheritance_model(int);

int            trait_inheritance_model_Q1(Estimate_Set *table, int model, int message)
{
  /* Local variables */
  int   df;

  /* Entry point for trait inheritance model */
  if (message) {
    print_inheritance_model(model);
  }

  switch (model) {
  case NO_ASSOCIATION_MODEL:
    no_association_model(table);
    df = 0;
    break;
  case DOMINANT_MODEL:
    dominant_model(table);
    df = 1;
    break;
  case RECESSIVE_MODEL:
    df = 1;
    recessive_model(table);
    break;
  case GENOTYPE_MODEL:
    genotype_model(table);
    df = 2;
    break;
  case OVER_DOMINANT_MODEL:
    over_dominant_model(table);
    df = 1;
    break;
  default:
    break;
  }

  return df;
}

int            no_association_model(Estimate_Set *table)
{
  /* Local variables */
  int             s, c;
  DiploClass     *class;

  for (s=0; s<table->number_of_set; s++) {
    class = (table->geno_info + s)->class;
    for (c=0; c<(table->geno_info + s)->number_of_combi; c++) {
      class->p_flag = (unsigned char)0;
      class++;
    }
  }

  return 0;
}

int            dominant_model(Estimate_Set *table)
{
  /* Local variables */
  int             s, c;
  DiploClass     *class;

  for (s=0; s<table->number_of_set; s++) {
    class = (table->geno_info + s)->class;
    for (c=0; c<(table->geno_info + s)->number_of_combi; c++) {
      class->p_flag = class->h_flag[0] | class->h_flag[1];
      class++;
    }
  }

  return 0;
}

int            recessive_model(Estimate_Set *table)
{
  /* Local variables */
  int             s, c;
  DiploClass     *class;

  for (s=0; s<table->number_of_set; s++) {
    class = (table->geno_info + s)->class;
    for (c=0; c<(table->geno_info + s)->number_of_combi; c++) {
      class->p_flag = class->h_flag[0] & class->h_flag[1];
      class++;
    }
  }

  return 0;
}

int            genotype_model(Estimate_Set *table)
{
  /* Local variables */
  int             s, c, count[3], RC;
  DiploClass     *class;

  count[0] = 0;
  count[1] = 0;
  count[2] = 0;
  for (s=0; s<table->number_of_set; s++) {
    class = (table->geno_info + s)->class;
    for (c=0; c<(table->geno_info + s)->number_of_combi; c++) {
      class->p_flag = class->h_flag[0] + class->h_flag[1];
      count[(int)class->p_flag]++;
      class++;
    }
  }

  RC = 0;
  if (count[0] == 0) {
    RC = RC+1;
  }
  if (count[1] == 0) {
    RC = RC+1;
  }
  if (count[2] == 0) {
    RC = RC+1;
  }

  return RC;
}

int            over_dominant_model(Estimate_Set *table)
{
  /* Local variables */
  int             s, c;
  DiploClass     *class;

  for (s=0; s<table->number_of_set; s++) {
    class = (table->geno_info + s)->class;
    for (c=0; c<(table->geno_info + s)->number_of_combi; c++) {
      class->p_flag = class->d_flag;
      class++;
    }
  }

  return 0;
}

int            print_inheritance_model(int model)
{
  /* Output trait inheritance model */

  switch (model) {
  case NO_ASSOCIATION_MODEL:
    printf("\n===== No association model\n");
    break;
  case DOMINANT_MODEL:
    printf("\n===== Dominant model\n");
    break;
  case RECESSIVE_MODEL:
    printf("\n===== Recessive model\n");
    break;
  case GENOTYPE_MODEL:
    printf("\n===== Genotype model\n");
    break;
  case OVER_DOMINANT_MODEL:
    printf("\n===== Over dominant model\n");
    break;
  default:
    break;
  }

  return 0;
}
