#include <stdio.h>
#include <stdlib.h>

void *mem_alloc(size_t, const char *);
void *mem_realloc(void *, size_t, const char *);
void  sort_int_array(int, unsigned int *); 
void  sort_reverse_index_of_double_array(int, double *, unsigned int *);
void  hpsort_reverse_index_of_double_array(int, double *, unsigned int *);
void  hpsort_index_of_double_array(int, double *, unsigned int *);
void  hpsort_index_of_int_array(int, unsigned int *, unsigned int *);

void *mem_alloc(size_t size, const char *message)
{
  void *ptr;

  ptr = malloc(size);
  if (ptr == NULL) {
    printf("\n*** %s was not allocated.\n",message);
    exit(1);
  }

  return ptr;
}

void *mem_realloc(void *ptr, size_t size, const char *message)
{
  ptr = realloc(ptr, size);
  if (ptr == NULL) {
    printf("\n*** %s was not allocated.\n",message);
    exit(1);
  }

  return ptr;
}

void sort_int_array(int size, unsigned int *iarray)
{
  unsigned int i, j, tmp;

  for (i=0; i<size; i++) {
    for (j=i; j<size; j++) {
      if (*(iarray+i) > *(iarray+j)) {
	tmp = *(iarray+i);
	*(iarray+i) = *(iarray+j);
	*(iarray+j) = tmp;
      }
    }
  }

  return;
}

void sort_reverse_index_of_double_array(int size, double *array,
					unsigned int *index)
{
  unsigned int i, j, tmp;

  for (i=0; i<size; i++) {
    *(index+i) = i;
  }

  for (i=0; i<size; i++) {
    for (j=i; j<size; j++) {
      if (*(array + *(index+i)) < *(array + *(index+j))) {
	tmp        = *(index+i);
	*(index+i) = *(index+j);
	*(index+j) = tmp;
      }
    }
  }

  return;
}

void hpsort_reverse_index_of_double_array(int size, double *array,
					  unsigned int *index)
{
  unsigned int      i, r, j, l, k, tmp;

  for (i=0; i<size; i++) {
    *(index + i) = i;
  }

  index--;

  if (size < 2) {
    return;
  }

  l = (size >> 1) + 1;
  r = size;

  while (1) {
    if (l > 1) {
      l--;
      tmp = *(index + l);
    }
    else {
      tmp = *(index + r);
      *(index + r) = *(index + 1);
      if (--r == 1) {
	*(index + 1) = tmp;
	return;                                     /* this is exit !!! */
      }
    }

    i = l;
    j = l << 1;
    while (j <= r) {
      if (j < r && (*(array + *(index + j)) > *(array + *(index + j + 1)))) {
	j++;
      }

      if (*(array + tmp) > *(array + *(index + j))) {
	*(index + i) = *(index + j);
	i = j;
	j = j << 1;
      }
      else {
	j = r + 1;
      }
    }
    *(index + i) = tmp;
  }
}

void hpsort_index_of_double_array(int size, double *array,
				  unsigned int *index)
{
  unsigned int      i, r, j, l, k, tmp;

  for (i=0; i<size; i++) {
    *(index + i) = i;
  }

  index--;

  if (size < 2) {
    return;
  }

  l = (size >> 1) + 1;
  r = size;

  while (1) {
    if (l > 1) {
      l--;
      tmp = *(index + l);
    }
    else {
      tmp = *(index + r);
      *(index + r) = *(index + 1);
      if (--r == 1) {
	*(index + 1) = tmp;
	return;                                     /* this is exit !!! */
      }
    }

    i = l;
    j = l << 1;
    while (j <= r) {
      if (j < r && (*(array + *(index + j)) < *(array + *(index + j + 1)))) {
	j++;
      }

      if (*(array + tmp) < *(array + *(index + j))) {
	*(index + i) = *(index + j);
	i = j;
	j = j << 1;
      }
      else {
	j = r + 1;
      }
    }
    *(index + i) = tmp;
  }
}

void hpsort_index_of_int_array(int size,
			       unsigned int *array,
			       unsigned int *index)
{
  unsigned int      i, r, j, l, k, tmp;

  for (i=0; i<size; i++) {
    *(index + i) = i;
  }

  index--;

  if (size < 2) {
    return;
  }

  l = (size >> 1) + 1;
  r = size;

  while (1) {
    if (l > 1) {
      l--;
      tmp = *(index + l);
    }
    else {
      tmp = *(index + r);
      *(index + r) = *(index + 1);
      if (--r == 1) {
	*(index + 1) = tmp;
	return;                                     /* this is exit !!! */
      }
    }

    i = l;
    j = l << 1;
    while (j <= r) {
      if (j < r && (*(array + *(index + j)) < *(array + *(index + j + 1)))) {
	j++;
      }

      if (*(array + tmp) < *(array + *(index + j))) {
	*(index + i) = *(index + j);
	i = j;
	j = j << 1;
      }
      else {
	j = r + 1;
      }
    }
    *(index + i) = tmp;
  }
}
