#! /bin/bash
# Written by Dmitri Zaykin
set -o nounset +o noclobber
R_exec='R --vanilla --quiet'

if [ $# != 4 ]; then
    echo "Need to supply arguments in the following order -"
    echo " " $0 Options_File Data_File Phenotype_File Number_Of_Permutations
    exit 1
fi

OptFile=$1
DatFile=$2
PheFile=$3
NumPerm=$4

PvalueFile=$1_$2_$3_pvals.txt
tmpFile=$PvalueFile.tmp.tmp
cp /dev/null $PvalueFile

# Get the 3 original p-values
declare -i i=0 n=0
while [ $i -le $n ]; do
    res=$(./varS -f $OptFile -g $DatFile -Q $PheFile | grep p_value | awk '{print $3}')
    ok=$(echo $res | awk '{print NF}')
    if [ $ok -eq '3' ]; then
	echo $res >> $PvalueFile
	let i+=1
    fi
done

echo "Asymptotic p-values:"
echo $res | awk '{printf("%f %f %f\n", $1,$2,$3)}'

# Get the permuted p-values
declare -i i=1 n=$NumPerm
while [ $i -le $n ]; do
    Seed=$RANDOM
    ./perm.x $Seed < $PheFile > $tmpFile
    res=$(./varS -f $OptFile -g $DatFile -Q $tmpFile | grep p_value | awk '{print $3}')
    ok=$(echo $res | awk '{print NF}')
    if [ $ok -eq '3' ]; then
	echo $res >> $PvalueFile
	let i+=1
    fi
done

/bin/rm $tmpFile

# this 'bash' function feeds R commannds directly into R
runR() {
$R_exec <<-Done
    p <- matrix(scan("$PvalueFile", quiet=TRUE), byrow=T, ncol=3)
    p01 <- (p[,1])[1]; pi1 <- (p[,1])[-1]
    p02 <- (p[,2])[1]; pi2 <- (p[,2])[-1]
    p03 <- (p[,3])[1]; pi3 <- (p[,3])[-1]
    pv1 <- length(pi1[pi1 <= p01]) / length(pi1)
    pv2 <- length(pi2[pi2 <= p02]) / length(pi2)
    pv3 <- length(pi3[pi3 <= p03]) / length(pi3)
    cat("P_values:", pv1, pv2, pv3, "\n")
    q()
Done
}

ppv=$(runR | grep -v cat | grep P_values | awk '{print $2,$3,$4}')
echo "Permutational p-values:"
echo $ppv
